# https://www.immersivelimit.com/tutorials/create-coco-annotations-from-scratch/#coco-dataset-format
from __future__ import annotations

import json
import logging
import os
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union, cast

import numpy as np
import numpy.typing as npt
from PIL import Image
from segments.typing import SegmentsDatasetCategory
from segments.utils import SEGMENTS_ORANGE, get_semantic_bitmap
from skimage import img_as_ubyte
from skimage.measure import regionprops
from tqdm import tqdm


# https://adamj.eu/tech/2021/05/13/python-type-hints-how-to-fix-circular-imports/
if TYPE_CHECKING:
    from segments.dataset import SegmentsDataset


#############
# Variables #
#############
RGB = Tuple[int, int, int]
RGBA = Tuple[int, int, int, int]
ColorMap = Union[List[RGBA], List[RGB]]
logger = logging.getLogger(__name__)
COLORMAP: ColorMap = [
    (0, 113, 188, 255),
    (216, 82, 24, 255),
    (236, 176, 31, 255),
    (125, 46, 141, 255),
    (118, 171, 47, 255),
    (76, 189, 237, 255),
    (161, 19, 46, 255),
    (255, 0, 0, 255),
    (255, 127, 0, 255),
    (190, 190, 0, 255),
    (0, 255, 0, 255),
    (0, 0, 255, 255),
    (170, 0, 255, 255),
    (84, 84, 0, 255),
    (84, 170, 0, 255),
    (84, 255, 0, 255),
    (170, 84, 0, 255),
    (170, 170, 0, 255),
    (170, 255, 0, 255),
    (255, 84, 0, 255),
    (255, 170, 0, 255),
    (255, 255, 0, 255),
    (0, 84, 127, 255),
    (0, 170, 127, 255),
    (0, 255, 127, 255),
    (84, 0, 127, 255),
    (84, 84, 127, 255),
    (84, 170, 127, 255),
    (84, 255, 127, 255),
    (170, 0, 127, 255),
    (170, 84, 127, 255),
    (170, 170, 127, 255),
    (170, 255, 127, 255),
    (255, 0, 127, 255),
    (255, 84, 127, 255),
    (255, 170, 127, 255),
]


##################
# Helper methods #
##################
# https://github.com/cocodataset/panopticapi/blob/master/panopticapi/utils.py
class IdGenerator:
    """
    The class is designed to generate unique IDs that have meaningful RGB encoding.
    Given semantic category unique ID will be generated and its RGB encoding will
    have color close to the predefined semantic category color.
    The RGB encoding used is ID = R * 256 * G + 256 * 256 + B.
    Class constructor takes dictionary ``{id: category_info}``, where all semantic
    class ids are presented and ``category_info`` record is a dict with fields
    ``isthing`` and ``color``
    """

    def __init__(self, categories: Dict[int, SegmentsDatasetCategory]):
        self.taken_colors: Set[RGB] = set()
        self.taken_colors.add((0, 0, 0))
        self.categories = categories
        for category in self.categories.values():
            if category.isthing == 0:
                self.taken_colors.add(category.color)

    def get_color(self, cat_id: int) -> RGB:
        def random_color(base: RGB, max_dist: int = 30) -> RGB:
            new_color: npt.NDArray[Any] = base + np.random.randint(low=-max_dist, high=max_dist + 1, size=3)
            rgb = tuple(np.maximum(0, np.minimum(255, new_color)))
            return cast(RGB, rgb)

        category = self.categories[cat_id]
        if category.isthing == 0:
            return category.color
        base_color = category.color
        if base_color not in self.taken_colors:
            self.taken_colors.add(base_color)
            return base_color
        else:
            while True:
                color = random_color(base_color)
                if color not in self.taken_colors:
                    self.taken_colors.add(color)
                    return color

    def get_id(self, cat_id: int) -> int:
        color = self.get_color(cat_id)
        return rgb2id(color)

    def get_id_and_color(self, cat_id: int) -> Tuple[int, RGB]:
        color = self.get_color(cat_id)
        return rgb2id(color), color


def rgb2id(color: Union[npt.NDArray[Any], RGB]) -> Union[npt.NDArray[Any], int]:
    """Convert rgb to an id.

    Args:
        color: An RGB value.

    Returns:
        The id.
    """
    if isinstance(color, np.ndarray) and len(color.shape) == 3:
        if color.dtype == np.uint8:
            color = color.astype(np.int32)
        color_id = color[:, :, 0] + 256 * color[:, :, 1] + 256 * 256 * color[:, :, 2]
        return cast(npt.NDArray[Any], color_id)
    return int(color[0] + 256 * color[1] + 256 * 256 * color[2])


def id2rgb(id_map: npt.NDArray[Any]) -> Union[npt.NDArray[Any], RGB]:
    """Convert a color id to an rgb.

    Args:
        id_map: An id map.

    Returns:
        An rgb.
    """
    if isinstance(id_map, np.ndarray):
        id_map_copy = id_map.copy()
        rgb_shape = tuple(list(id_map.shape) + [3])
        rgb_map: npt.NDArray[Any] = np.zeros(rgb_shape, dtype=np.uint8)
        for i in range(3):
            rgb_map[..., i] = id_map_copy % 256
            id_map_copy //= 256
        return rgb_map
    color = []
    for _ in range(3):
        color.append(id_map % 256)
        id_map //= 256
    return tuple(color)


def get_color(id: int) -> RGB:
    id = id % len(COLORMAP)
    return COLORMAP[id][0:3]


def colorize(img: npt.NDArray[Any], colormap: Optional[ColorMap] = None) -> npt.NDArray[Any]:
    indices = np.unique(img)
    indices = indices[indices != 0]

    colored_img = np.zeros((img.shape[0], img.shape[1], 3), np.uint8)

    for id in indices:
        mask = img == id
        if colormap:
            color = colormap[id - 1]
        else:
            color = get_color(id - 1)
        colored_img[mask] = color

    return colored_img


def get_bbox(binary_mask: npt.NDArray[Any]) -> Union[Tuple[int, int, int, int], bool]:
    """Returns the bounding box of the binary mask (if one is found, otherwise returns False)

    Args:
        binary_mask: The binary mask as a numpy array.

    Returns:
        The bounding box or False if no bounding box is found.
    """

    regions = regionprops(np.uint8(binary_mask))
    if len(regions) == 1:
        bbox = regions[0].bbox
        return cast(Tuple[int, int, int, int], bbox)

    return False


##################
# Export methods #
##################
def export_coco_instance(dataset: SegmentsDataset, export_folder: str) -> Tuple[str, Optional[str]]:
    """Export a Segments dataset as a coco instance.

    Args:
        dataset: A :class:`.SegmentsDataset`.
        export_folder: The output directory.

    Returns:
        Returns the file name and the image directory name.

    Raises:
        :exc:`ImportError`: If pycocotools is not installed.
    """
    try:
        from pycocotools import mask as pctmask
    except ImportError as e:
        logger.error(
            "Please install pycocotools first: pip install pycocotools. Or on Windows: pip install pycocotools-windows"
        )
        raise e

    # Create export folder
    os.makedirs(export_folder, exist_ok=True)

    info = {
        "description": dataset.release["dataset"]["name"],
        "version": dataset.release["name"],
        # 'year': 2022,
    }

    categories = dataset.categories
    task_type = dataset.task_type
    # for i, category in enumerate(dataset.project_info['label_taxonomy']):
    #     categories.append({
    #         'id': i+1,
    #         'supercategory': 'object',
    #         'name': category
    #     })

    images = []
    annotations = []

    annotation_id = 1
    for i in tqdm(range(len(dataset)), total=len(dataset), colour=SEGMENTS_ORANGE):
        sample = dataset[i]

        if sample["annotations"] is None:
            continue

        image_id = i + 1
        images.append(
            {
                "id": image_id,
                "file_name": sample["file_name"],
                "height": sample["image"].size[1] if sample["image"] else None,
                "width": sample["image"].size[0] if sample["image"] else None,
            }
        )

        if task_type == "segmentation-bitmap" or task_type == "segmentation-bitmap-highres":
            # https://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.regionprops
            regions = regionprops(np.array(sample["segmentation_bitmap"], np.uint32))
            regions = {region.label: region for region in regions}

        for instance in sample["annotations"]:
            category_id = instance["category_id"]

            annotation = {
                "id": annotation_id,
                "image_id": image_id,
                "category_id": category_id,
            }

            # Segmentation bitmap labels
            if task_type == "segmentation-bitmap" or task_type == "segmentation-bitmap-highres":
                if instance["id"] not in regions:
                    # Only happens when the instance has 0 labeled pixels, which should not happen.
                    logger.warning(
                        f"Skipping instance with 0 labeled pixels: {sample['file_name']}, instance_id: {instance['id']}, category_id: {category_id}"
                    )
                    continue

                instance_mask = np.array(sample["segmentation_bitmap"], np.uint32) == instance["id"]

                region = regions[instance["id"]]
                bbox = region.bbox
                # bbox = get_bbox(instance_mask)

                y0, x0, y1, x1 = bbox
                # rle = mask.encode(np.asfortranarray(instance_mask))
                rle = pctmask.encode(np.array(instance_mask[:, :, None], dtype=np.uint8, order="F"))[
                    0
                ]  # https://github.com/matterport/Mask_RCNN/issues/387#issuecomment-522671380
                #         instance_mask_crop = instance_mask[y0:y1, x0:x1]
                #         rle = mask.encode(np.asfortranarray(instance_mask_crop))
                #         plt.imshow(instance_mask_crop)
                #         plt.show()

                # area = int(mask.area(rle))
                area = int(region.area)
                rle["counts"] = rle["counts"].decode("ascii")

                annotation.update(
                    {
                        "bbox": [x0, y0, x1 - x0, y1 - y0],
                        #             'bbox_mode': BoxMode.XYWH_ABS,
                        "segmentation": rle,
                        "area": area,
                        "iscrowd": 0,
                    }
                )

            # Vector labels
            elif "type" in instance:
                # bbox
                if instance["type"] == "bbox":
                    points = instance["points"]
                    x0 = points[0][0]
                    y0 = points[0][1]
                    x1 = points[1][0]
                    y1 = points[1][1]

                    annotation.update({"bbox": [x0, y0, x1 - x0, y1 - y0]})

                # keypoints
                elif instance["type"] == "point":
                    points = instance["points"]
                    x0 = points[0][0]
                    y0 = points[0][1]

                    annotation.update(
                        {
                            "keypoints": [
                                x0,
                                y0,
                                1,
                            ],  # https://cocodataset.org/#format-results
                        }
                    )

                # polygon
                elif instance["type"] == "polygon":
                    annotation.update({"points": instance["points"]})

                # polyline
                elif instance["type"] == "polyline":
                    logger.warning("Polyline annotations are not exported.")

                else:
                    assert False

            annotations.append(annotation)
            annotation_id += 1

    json_data = {
        "info": info,
        "categories": [category.model_dump(mode="json") for category in categories],
        "images": images,
        "annotations": annotations,
        # "segment_info": [] # Only in Panoptic annotations
    }

    file_name = os.path.join(
        export_folder,
        f"export_coco-instance_{dataset.dataset_identifier}_{dataset.release['name']}.json",
    )
    with open(file_name, "w") as f:
        json.dump(json_data, f)

    print(f"Exported to {file_name}. Images in {dataset.image_dir}")
    return file_name, dataset.image_dir


def export_coco_panoptic(dataset: SegmentsDataset, export_folder: str) -> Tuple[str, Optional[str]]:
    """Export a Segments dataset in COCO panoptic format.

    Args:
        dataset: A :class:`.SegmentsDataset`.
        export_folder: The output directory.

    Returns:
        Returns the file name and the image directory name.
    """

    # Create export folder
    os.makedirs(export_folder, exist_ok=True)

    # INFO
    info = {
        "description": dataset.release["dataset"]["name"],
        "version": dataset.release["name"],
        # 'year': '2022'
    }

    # CATEGORIES
    categories: List[SegmentsDatasetCategory] = []
    for i, category in enumerate(dataset.categories):
        color = category.color[:3] if category.color else get_color(i)
        isthing = int(category.has_instances) if hasattr(category, "has_instances") else 0

        categories.append(
            SegmentsDatasetCategory(
                id=category.id,
                name=category.name,
                color=color,
                isthing=isthing,
            )
        )

    categories_dict: Dict[int, SegmentsDatasetCategory] = {category.id: category for category in categories}
    id_generator = IdGenerator(categories_dict)

    # IMAGES AND ANNOTATIONS
    images = []
    annotations = []
    for i in tqdm(range(len(dataset)), total=len(dataset), colour=SEGMENTS_ORANGE):
        sample = dataset[i]

        if sample["annotations"] is None:
            continue

        # Images
        image_id = i + 1
        images.append(
            {
                "id": image_id,
                "file_name": sample["file_name"],
                "height": sample["image"].size[1] if sample["image"] else None,
                "width": sample["image"].size[0] if sample["image"] else None,
            }
        )

        # Annotations
        panoptic_label = np.zeros(
            (
                sample["segmentation_bitmap"].size[1],
                sample["segmentation_bitmap"].size[0],
                3,
            ),
            np.uint8,
        )

        segments_info = []

        # https://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.regionprops
        regions = regionprops(np.array(sample["segmentation_bitmap"], np.uint32))
        regions = {region.label: region for region in regions}

        for instance in sample["annotations"]:
            category_id = instance["category_id"]

            instance_id, color = id_generator.get_id_and_color(category_id)

            if instance["id"] not in regions:
                # Only happens when the instance has 0 labeled pixels, which should not happen.
                logger.warning(
                    f"Skipping instance with 0 labeled pixels: {sample['file_name']}, instance_id: {instance['id']}, category_id: {category_id}"
                )
                continue

            # Read the instance mask and fill in the panoptic label. TODO: take this out of the loop to speed things up.
            instance_mask = np.array(sample["segmentation_bitmap"], np.uint32) == instance["id"]
            panoptic_label[instance_mask] = color

            # bbox = get_bbox(instance_mask)
            region = regions[instance["id"]]
            bbox = region.bbox
            y0, x0, y1, x1 = bbox

            # rle = mask.encode(np.array(instance_mask[:,:,None], dtype=np.uint8, order='F'))[0] # https://github.com/matterport/Mask_RCNN/issues/387#issuecomment-522671380
            # area = int(mask.area(rle))
            area = int(region.area)

            segments_info.append(
                {
                    "id": instance_id,
                    "category_id": category_id,
                    "bbox": [x0, y0, x1 - x0, y1 - y0],
                    "area": area,
                    "iscrowd": 0,
                }
            )

        file_name = os.path.splitext(os.path.basename(sample["name"]))[0]
        label_file_name = f"{file_name}_label_{dataset.labelset}_coco-panoptic.png"
        annotations.append(
            {
                "segments_info": segments_info,
                "file_name": label_file_name,
                "image_id": image_id,
            }
        )

        # # Image
        # image = sample['image']
        # export_file = os.path.join(label_export_folder, '{}.png'.format(file_name))
        # image.save(export_file)

        # # Instance png
        # instance_label = sample['segmentation_bitmap']
        # export_file = os.path.join(dataset.image_dir, '{}_label_{}_instance.png'.format(file_name, dataset.labelset))
        # instance_label.save(export_file)

        # # Colored instance png
        # instance_label_colored = colorize(np.uint8(instance_label))
        # export_file = os.path.join(dataset.image_dir, '{}_label{}_instance_colored.png'.format(file_name, dataset.labelset))
        # Image.fromarray(img_as_ubyte(instance_label_colored)).save(export_file)

        # Panoptic png
        export_file = os.path.join(dataset.image_dir, label_file_name)
        Image.fromarray(panoptic_label).save(export_file)

        # # Semantic png
        # semantic_label = get_semantic_bitmap(instance_label, sample['annotations'])
        # export_file = os.path.join(dataset.image_dir, '{}_label_{}_semantic.png'.format(file_name, dataset.labelset))
        # Image.fromarray(img_as_ubyte(semantic_label)).save(export_file)

        # # Colored semantic png
        # semantic_label_colored = colorize(np.uint8(semantic_label), colormap=[c['color'] for c in categories])
        # export_file = os.path.join(dataset.image_dir, '{}_label_{}_semantic_colored.png'.format(file_name, dataset.labelset))
        # Image.fromarray(img_as_ubyte(semantic_label_colored)).save(export_file)

    # PUT EVERYTHING TOGETHER
    json_data = {
        "info": info,
        "categories": [category.model_dump(mode="json") for category in categories],
        "images": images,
        "annotations": annotations,
    }

    # WRITE JSON TO FILE
    file_name = os.path.join(
        export_folder,
        f"export_coco-panoptic_{dataset.dataset_identifier}_{dataset.release['name']}.json",
    )
    with open(file_name, "w") as f:
        json.dump(json_data, f)

    print(f"Exported to {file_name}. Images and labels in {dataset.image_dir}")
    return file_name, dataset.image_dir


def export_image(
    dataset: SegmentsDataset,
    export_folder: str,
    export_format: str,
    id_increment: int,
) -> Optional[str]:
    """Export a Segments dataset as images.

    Args:
        dataset: A :class:`.SegmentsDataset`.
        export_folder: The output directory.
        export_format: The export format.
        id_increment: Increment the category ids with this number.

    Returns:
        Returns the export folder.
    """
    # Create export folder

    if export_folder == ".":
        export_folder = dataset.image_dir

    os.makedirs(export_folder, exist_ok=True)

    # CATEGORIES
    categories = []
    for i, category in enumerate(dataset.categories):
        color = category.color[:3] if category.color else get_color(i)
        isthing = int(category.has_instances) if hasattr(category, "has_instances") else 0

        categories.append(
            SegmentsDatasetCategory.model_validate(
                {
                    "id": category.id,
                    "name": category.name,
                    "color": color,
                    "isthing": isthing,
                }
            )
        )

    for i in tqdm(range(len(dataset)), total=len(dataset), colour=SEGMENTS_ORANGE):
        sample = dataset[i]

        if sample["annotations"] is None:
            continue

        file_name = os.path.splitext(os.path.basename(sample["name"]))[0]

        # # Image
        # image = sample['image']
        # export_file = os.path.join(label_export_folder, '{}.png'.format(file_name))
        # image.save(export_file)

        if export_format == "instance":
            # Instance png
            instance_label = sample["segmentation_bitmap"]
            export_file = os.path.join(
                export_folder,
                f"{file_name}_label_{dataset.labelset}_instance.png",
            )
            instance_label.save(export_file)

        elif export_format == "instance-color":
            # Colored instance png
            instance_label = sample["segmentation_bitmap"]
            instance_label_colored = colorize(np.uint8(instance_label))
            export_file = os.path.join(
                export_folder,
                f"{file_name}_label_{dataset.labelset}_instance_colored.png",
            )
            Image.fromarray(img_as_ubyte(instance_label_colored)).save(export_file)

        elif export_format == "semantic":
            # Semantic png
            instance_label = sample["segmentation_bitmap"]
            semantic_label = get_semantic_bitmap(instance_label, sample["annotations"], id_increment)
            export_file = os.path.join(
                export_folder,
                f"{file_name}_label_{dataset.labelset}_semantic.png",
            )
            Image.fromarray(img_as_ubyte(semantic_label)).save(export_file)

        elif export_format == "semantic-color":
            # Colored semantic png
            instance_label = sample["segmentation_bitmap"]
            semantic_label = get_semantic_bitmap(instance_label, sample["annotations"], id_increment)
            semantic_label_colored = colorize(np.uint8(semantic_label), colormap=[c.color for c in categories])
            export_file = os.path.join(
                export_folder,
                f"{file_name}_label_{dataset.labelset}_semantic_colored.png",
            )
            Image.fromarray(img_as_ubyte(semantic_label_colored)).save(export_file)

    print(f"Exported to {export_folder}")
    return export_folder


def export_yolo(
    dataset: SegmentsDataset,
    export_folder: str,
    image_width: Optional[float] = None,
    image_height: Optional[float] = None,
) -> Optional[str]:
    """Export a Segments dataset in YOLO format.

    Args:
        dataset: A segments dataset.
        export_folder: The output directory.
        image_width: The width of the image (needed for ``image-vector-sequence``).
        image_height: The height of the image (needed for ``image-vector-sequence``).

    Returns:
        The image directory of the dataset.

    Raises:
        :exc:`ValueError`: If the dataset is not a bounding box dataset.
        :exc:`ValueError`: If the dataset is an ``image-vector-sequence``and the image width or image height is :obj:`None`.
    """

    def write_yolo_file(file_name: str, annotations: Any, image_width: float, image_height: float) -> None:
        with open(file_name, "w") as f:
            for annotation in annotations:
                if annotation["type"] == "bbox":
                    category_id = annotation["category_id"]
                    [[x0, y0], [x1, y1]] = annotation["points"]

                    # Normalize
                    x0, x1 = x0 / image_width, x1 / image_width
                    y0, y1 = y0 / image_height, y1 / image_height

                    # Get center, width and height of bbox
                    x_center = (x0 + x1) / 2
                    y_center = (y0 + y1) / 2
                    width = abs(x1 - x0)
                    height = abs(y1 - y0)

                    # Save it to the file
                    # print(category_id, x_center, y_center, width, height)
                    f.write(f"{category_id} {x_center:.6f} {y_center:.6f} {width:.6f} {height:.6f}\n")

    # Create export folder
    os.makedirs(os.path.join(export_folder, dataset.image_dir), exist_ok=True)

    if dataset.task_type not in ["vector", "bboxes", "image-vector-sequence"]:
        raise ValueError("You can only export bounding box datasets to YOLO format.")

    if dataset.task_type == "vector":
        logger.warning(
            "Only bounding box annotations will be processed. Polygon, polyline and keypoint annotations will be ignored."
        )

    if dataset.task_type == "image-vector-sequence":
        logger.warning(
            "Note that the sequences will be exported as individual frames, disregarding the tracking information."
        )
        for i in tqdm(range(len(dataset)), total=len(dataset), colour=SEGMENTS_ORANGE):
            sample = dataset[i]
            image_name = os.path.splitext(os.path.basename(sample["name"]))[0]

            if image_width is None or image_height is None:
                raise ValueError("Please provide image_width and image_height parameters (i.e., not None).")

            for j, frame in enumerate(sample["labels"]["ground-truth"]["attributes"]["frames"]):
                # Construct the file name from image and frame name
                try:
                    frame_name = sample["attributes"]["frames"][j]["name"]
                except (KeyError, TypeError):
                    frame_name = f"{j + 1:05d}"
                file_name = os.path.join(export_folder, dataset.image_dir, f"{image_name}-{frame_name}.txt")

                # Testing on x is the same as testing len(x)>0 (this also checks that x is not None - see truthy and falsy values in Python)
                # https://stackoverflow.com/questions/39983695/what-is-truthy-and-falsy-how-is-it-different-from-true-and-false
                if "annotations" in frame and frame["annotations"]:
                    annotations = frame["annotations"]
                    write_yolo_file(file_name, annotations, image_width, image_height)
    else:
        for i in tqdm(range(len(dataset)), total=len(dataset), colour=SEGMENTS_ORANGE):
            sample = dataset[i]
            image_name = os.path.splitext(os.path.basename(sample["name"]))[0]
            file_name = os.path.join(export_folder, dataset.image_dir, f"{image_name}.txt")

            if "annotations" in sample and sample["annotations"]:
                annotations = sample["annotations"]
                write_yolo_file(
                    file_name,
                    annotations,
                    cast(
                        float,
                        image_width if image_width is not None else sample["image"].width,
                    ),
                    cast(
                        float,
                        image_height if image_height is not None else sample["image"].height,
                    ),
                )

    print(f"Exported. Images and labels in {dataset.image_dir}")
    return dataset.image_dir


def export_polygon(dataset: SegmentsDataset, export_folder: str) -> Tuple[str, Optional[str]]:
    """Export a Segments dataset as polygons (i.e., contours).

    Args:
        dataset: A :class:`.SegmentsDataset`.
        export_folder: The output directory.

    Raises:
        :exc:`ImportError`: If OpenCV is not installed.

    Returns:
        Returns the file name and the image directory name.
    """
    try:
        import cv2 as cv
    except ImportError as e:
        logger.error("Please install OpenCV first: pip install opencv-python.")
        raise e

    # Create export folder
    os.makedirs(export_folder, exist_ok=True)

    categories = dataset.categories
    # task_type = dataset.task_type

    images = []
    annotations = []

    annotation_id = 1
    for i in tqdm(range(len(dataset)), total=len(dataset), colour=SEGMENTS_ORANGE):
        sample = dataset[i]

        if sample["annotations"] is None:
            continue

        image_id = i + 1
        images.append(
            {
                "id": image_id,
                "file_name": sample["file_name"],
                "height": sample["image"].size[1] if sample["image"] else None,
                "width": sample["image"].size[0] if sample["image"] else None,
            }
        )

        segmentation_bitmap = np.array(sample["segmentation_bitmap"], np.uint32)
        # https://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.regionprops
        regions = regionprops(segmentation_bitmap)

        regions = {region.label: region for region in regions}

        segmentation_bitmap_uint8 = segmentation_bitmap.astype(np.uint8)

        for instance in sample["annotations"]:
            category_id = instance["category_id"]

            annotation = {
                "id": annotation_id,
                "image_id": image_id,
                "category_id": category_id,
                "attributes": instance.get("attributes", None),
            }

            if instance["id"] not in regions:
                # Only happens when the instance has 0 labeled pixels, which should not happen.
                logger.warning(
                    f"Skipping instance with 0 labeled pixels: {sample['file_name']}, instance_id: {instance['id']}, category_id: {category_id}"
                )
                continue

            instance_mask = segmentation_bitmap_uint8 == instance["id"]

            # Allowed OpenCV data types: https://stackoverflow.com/questions/12785121/access-opencv-matrix-cv-32s-element
            black_white_image = instance_mask.astype("uint8")

            # Contour retrieval modes (all, outer, in a hierarchy, etc.): https://docs.opencv.org/4.x/d3/dc0/group__imgproc__shape.html#ga819779b9857cc2f8601e6526a3a5bc71
            contours, hierarchy = cv.findContours(black_white_image, cv.RETR_LIST, cv.CHAIN_APPROX_SIMPLE)

            # Make object json serializable
            serializable_contours = [np.squeeze(contour).tolist() for contour in contours]

            annotation.update({"polygons": serializable_contours})

            annotations.append(annotation)
            annotation_id += 1

    json_data = {
        "categories": [category.model_dump(mode="json") for category in categories],
        "images": images,
        "annotations": annotations,
    }

    file_name = os.path.join(
        export_folder,
        f"export_polygon_{dataset.dataset_identifier}_{dataset.release['name']}.json",
    )
    with open(file_name, "w") as f:
        json.dump(json_data, f)

    print(f"Exported to {file_name}. Images in {dataset.image_dir}")
    return file_name, dataset.image_dir



import time
import typing
from typing import Final

import pydantic
import pytest
import segments
import segments.typing as segments_typing
from segments.client import SegmentsClient
from segments.exceptions import AlreadyExistsError, CollaboratorError, InvalidModelError
from segments.resource_api import Collaborator, Dataset, Issue, Label, Labelset, Release, Sample
from segments.typing import (
    LabelStatus,
    PointcloudSegmentationLabelAttributes,
    Role,
    TaskType,
)


@pytest.mark.parametrize(
    "func1,func2,ignore",
    [
        (Dataset.add_sample, SegmentsClient.add_sample, ["dataset_identifier"]),
        (Dataset.delete, SegmentsClient.delete_dataset, ["dataset_identifier"]),
        (Dataset.update, SegmentsClient.update_dataset, ["dataset_identifier"]),
        (Dataset.get_samples, SegmentsClient.get_samples, ["dataset_identifier"]),
        (Dataset.get_collaborator, SegmentsClient.get_dataset_collaborator, ["dataset_identifier"]),
        (Dataset.add_collaborator, SegmentsClient.add_dataset_collaborator, ["dataset_identifier"]),
        (Dataset.add_release, SegmentsClient.add_release, ["dataset_identifier"]),
        (Dataset.get_release, SegmentsClient.get_release, ["dataset_identifier"]),
        (Dataset.get_releases, SegmentsClient.get_releases, ["dataset_identifier"]),
        (Dataset.get_labelsets, SegmentsClient.get_labelsets, ["dataset_identifier"]),
        (Dataset.get_labelset, SegmentsClient.get_labelset, ["dataset_identifier"]),
        (Dataset.add_labelset, SegmentsClient.add_labelset, ["dataset_identifier"]),
        (Dataset.get_issues, SegmentsClient.get_issues, ["dataset_identifier"]),
        (Dataset.get_workunits, SegmentsClient.get_workunits, ["dataset_identifier"]),
        (Dataset.clone, SegmentsClient.clone_dataset, ["dataset_identifier"]),
        (Sample.delete, SegmentsClient.delete_sample, ["uuid"]),
        (Sample.update, SegmentsClient.update_sample, ["uuid"]),
        (Sample.get_label, SegmentsClient.get_label, ["sample_uuid"]),
        (Sample.add_label, SegmentsClient.add_label, ["sample_uuid"]),
        (Sample.update_label, SegmentsClient.update_label, ["sample_uuid"]),
        (Sample.delete_label, SegmentsClient.delete_label, ["sample_uuid"]),
        (Sample.add_issue, SegmentsClient.add_issue, ["sample_uuid"]),
        (Label.update, SegmentsClient.update_label, ["sample_uuid", "labelset"]),
        (Label.delete, SegmentsClient.delete_label, ["sample_uuid", "labelset"]),
        (Issue.delete, SegmentsClient.delete_issue, ["uuid"]),
        (Issue.update, SegmentsClient.update_issue, ["uuid"]),
        (Collaborator.update, SegmentsClient.update_dataset_collaborator, ["dataset_identifier", "username"]),
        (Collaborator.delete, SegmentsClient.delete_dataset_collaborator, ["dataset_identifier", "username"]),
        (Labelset.delete, SegmentsClient.delete_labelset, ["dataset_identifier", "name"]),
        (Release.delete, SegmentsClient.delete_release, ["dataset_identifier", "name"]),
    ],
)
def test_same_arguments(func1, func2, ignore: list[str]):
    sig1 = typing.get_type_hints(func1)
    sig2 = typing.get_type_hints(func2)

    ignored_arguments_set = set(ignore)
    func1_argnames = sig1.keys() - ignored_arguments_set
    func2_argnames = sig2.keys() - ignored_arguments_set
    if func1_argnames != func2_argnames:
        raise AssertionError(
            f"Arguments in {func1.__name__} and {func2.__name__} are not the same: {func1_argnames ^ func2_argnames}"
        )


@pytest.mark.usefixtures("setup_class_client")
class TestDataset:
    def test_add_update_delete_dataset(self, add_update_dataset_arguments) -> None:
        arguments = add_update_dataset_arguments
        dataset = None
        try:
            # Add dataset
            dataset = self.client.add_dataset(**arguments)
            assert isinstance(dataset, Dataset)
            # Update dataset
            del arguments["name"]
            del arguments["organization"]
            dataset = dataset.update(**arguments)
            assert isinstance(dataset, Dataset)
        except AlreadyExistsError:
            pass
        finally:
            # Delete dataset
            if dataset is not None:
                dataset.delete()

    def test_clone_dataset_defaults(self, owner) -> None:
        name = "example-images-segmentation"
        dataset_identifier = f"{owner}/{name}"
        clone = None
        try:
            dataset = self.client.get_dataset(dataset_identifier)
            clone = dataset.clone()

            assert isinstance(clone, Dataset)
            assert clone.name == "example-images-segmentation-clone"
            assert clone.task_type == "segmentation-bitmap"
            assert not clone.public
        except AlreadyExistsError:
            pass
        finally:
            # Delete dataset
            if clone is not None:
                clone.delete()

    def test_clone_dataset_custom(self, owner) -> None:
        dataset_identifier = f"{owner}/example-images-vector"

        new_name = "example-images-vector-clone"
        new_task_type = TaskType.VECTOR
        clone = None

        try:
            dataset = self.client.get_dataset(dataset_identifier)
            clone = dataset.clone(
                new_name=new_name,
                new_task_type=new_task_type,
                organization=owner,
            )

            assert isinstance(clone, Dataset)
            assert clone.name == new_name
            assert clone.task_type == new_task_type
        except AlreadyExistsError:
            pass
        finally:
            # Delete dataset
            if clone is not None:
                clone.delete()

    def test_get_add_update_delete_dataset_collaborator(self, owner, datasets) -> None:
        username = "admin-arnaud"
        role = Role.ADMIN
        dataset_name = datasets[0]
        dataset_identifier = f"{owner}/{dataset_name}"

        dataset = self.client.get_dataset(dataset_identifier)
        collaborator = None
        try:
            # Add collaborator
            collaborator = dataset.add_collaborator(username, role)
            assert isinstance(collaborator, segments_typing.Collaborator)
            # Get collaborator
            collaborator_get = dataset.get_collaborator(username)
            assert isinstance(collaborator, segments_typing.Collaborator)
            assert collaborator == collaborator_get

            collaborator_update = collaborator.update(Role.ADMIN)
            assert collaborator_update.role == Role.ADMIN
        except CollaboratorError as e:
            # Collaborator already exists, clean it up and fail the test. This should let the test pass on the next run.
            self.client.delete_dataset_collaborator(dataset_identifier, username)
            raise e
        finally:
            # Delete collaborator
            if collaborator is not None:
                collaborator.delete()


@pytest.mark.usefixtures("setup_class_client")
class TestSample:
    def test_get_samples(self, datasets, owner) -> None:
        name = None
        label_status = LabelStatus.UNLABELED
        metadata = None
        sort: Final = "created"
        direction: Final = "desc"
        labelset = None  # "ground-truth"
        for dataset_name in datasets:
            dataset_identifier = f"{owner}/{dataset_name}"
            dataset = self.client.get_dataset(dataset_identifier)
            samples = dataset.get_samples(dataset_identifier, labelset, name, label_status, metadata, sort, direction)
            for sample in samples:
                assert isinstance(sample, Sample)

    def test_add_update_delete_sample(
        self, datasets, owner, sample_attribute_types, TIME_INTERVAL, add_update_sample_arguments
    ) -> None:
        attributes_dict = add_update_sample_arguments

        metadata = {"weather": "sunny", "camera_id": 3}
        priority = 0
        name = "Test sample"

        # check if test sample already exists, if so delete it
        for dataset_name in datasets:
            dataset = self.client.get_dataset(f"{owner}/{dataset_name}")
            samples = dataset.get_samples()
            if any(sample.name == "Test sample" for sample in samples):
                sample = next(sample for sample in samples if sample.name == "Test sample")
                sample.delete()

        for sample_attribute_type, dataset_name in zip(sample_attribute_types, datasets):
            dataset_identifier = f"{owner}/{dataset_name}"
            dataset = self.client.get_dataset(dataset_identifier)
            attributes = attributes_dict[sample_attribute_type]
            sample = None
            try:
                sample = dataset.add_sample(
                    name,
                    attributes,
                    metadata,
                    priority,
                )
                assert isinstance(sample, Sample)
                sample = sample.update(
                    name,
                    attributes,
                    metadata,
                    priority,
                )
                assert isinstance(sample, Sample)
            finally:
                if sample is not None:
                    time.sleep(TIME_INTERVAL)
                    sample.delete()

    def test_unset_assigned(self, datasets, owner, add_update_sample_arguments, sample_attribute_types) -> None:
        attributes = add_update_sample_arguments
        name = "test_unset_assigned"

        attributes_type = sample_attribute_types[0]
        dataset_name = datasets[0]
        dataset_identifier = f"{owner}/{dataset_name}"

        dataset = self.client.get_dataset(dataset_identifier)

        if matched_samples := dataset.get_samples(name=name):
            assert len(matched_samples) == 1, f"Multiple matches found for sample with name {name}"
            matched_samples[0].delete()

        sample = dataset.add_sample(name, attributes[attributes_type], assigned_reviewer=owner, assigned_labeler=owner)
        try:
            assert sample.assigned_labeler == owner
            assert sample.assigned_reviewer == owner

            sample = sample.update(assigned_labeler=None)

            assert sample.assigned_labeler is None
            assert sample.assigned_reviewer == owner

            sample = sample.update(assigned_reviewer=None)

            assert sample.assigned_reviewer is None
        finally:
            sample.delete()


@pytest.mark.usefixtures("setup_class_client")
class TestLabel:
    def test_add_update_get_delete_label(
        self, sample_uuids, label_attribute_types, label_test_attributes, TIME_INTERVAL
    ) -> None:
        labelset = "ground-truth"
        label_status = LabelStatus.PRELABELED
        score = 1
        for sample_uuid, label_attribute_type in zip(sample_uuids, label_attribute_types):
            attributes = label_test_attributes[label_attribute_type]
            sample = self.client.get_sample(sample_uuid)
            label = None
            try:
                # Add
                label = sample.add_label(labelset, attributes, label_status, score)
                assert isinstance(label, Label)
                # Update
                label = label.update(attributes, label_status, score)
                assert isinstance(label, Label)
                # Also test the sample.update_label interface
                label = sample.update_label(labelset, attributes, label_status, score)
                assert isinstance(label, Label)
                label_get = sample.get_label(labelset)
                assert label.attributes == label_get.attributes
            except AlreadyExistsError:
                pass
            finally:
                if label is not None:
                    # Delete
                    time.sleep(TIME_INTERVAL)
                    label.delete()

    def test_validation_error(self, sample_uuids, label_attribute_types, bad_image_segmentation_label, TIME_INTERVAL):
        sample_idx = label_attribute_types.index("image-segmentation")
        sample_uuid = sample_uuids[sample_idx]
        sample = self.client.get_sample(sample_uuid)
        label = None
        with pytest.raises(pydantic.ValidationError) as excinfo:
            try:
                label = sample.add_label("ground-truth", bad_image_segmentation_label)
            finally:
                if label is not None:
                    time.sleep(TIME_INTERVAL)
                    label.delete()

        # Convenience api should only throw one error
        assert excinfo.value.error_count() == 1

        with pytest.raises(segments.exceptions.ValidationError) as excinfo:
            try:
                label = self.client.add_label(sample.uuid, "ground-truth", bad_image_segmentation_label)
            finally:
                if label is not None:
                    time.sleep(TIME_INTERVAL)
                    label.delete()

        # We expect the client api to throw many more errors
        assert excinfo.value.cause.error_count() > 1

    def test_provide_wrong_model(self, sample_uuids, label_attribute_types, label_test_attributes):
        # We'll try to provide a pointcloud segmentation model for an image segmentation label
        sample_idx = label_attribute_types.index("image-segmentation")
        sample_uuid = sample_uuids[sample_idx]

        bad_attributes = label_test_attributes["pointcloud-segmentation"]

        sample = self.client.get_sample(sample_uuid)
        with pytest.raises(InvalidModelError):
            sample.add_label("ground-truth", PointcloudSegmentationLabelAttributes.model_validate(bad_attributes))

    def test_vector_label_attributes_contains_track_id(
        self, sample_uuids, label_attribute_types, label_test_attributes
    ):
        task_type = "pointcloud-sequence-vector"
        sample_idx = label_attribute_types.index(task_type)
        sample_uuid = sample_uuids[sample_idx]
        sample = self.client.get_sample(sample_uuid)

        try:
            sample.add_label(labelset="ground-truth", attributes=label_test_attributes[task_type])
            label = sample.get_label()

            for annotation in label.attributes.frames[0].annotations:
                assert isinstance(annotation.track_id, int)
        finally:
            sample.delete_label(labelset="ground-truth")


@pytest.mark.usefixtures("setup_class_client")
class TestIssue:
    def test_get_issues(self, datasets, owner) -> None:
        for dataset in datasets:
            dataset_identifier = f"{owner}/{dataset}"
            dataset = self.client.get_dataset(dataset_identifier)
            issues = dataset.get_issues()
            for issue in issues:
                assert isinstance(issue, segments_typing.Issue)

    def test_add_update_delete_issue(self, sample_uuids) -> None:
        description = "You forgot to label this car."
        anchor = {"type": "location", "frame": 0, "sensor": "Lidar", "location": {"type": "2d", "x": 1.2, "y": 3.5}}
        sample_uuid = sample_uuids[0]
        sample = self.client.get_sample(sample_uuid)
        issue = None
        try:
            # Add issue
            issue = sample.add_issue(description, anchor=anchor)
            assert isinstance(issue, segments_typing.Issue)
            description2 = description.replace("car", "bike")
            issue = issue.update(description2, anchor=anchor)
            assert isinstance(issue, segments_typing.Issue)
            assert issue.description == description2
        finally:
            # Delete issue
            if issue:
                issue.delete()


@pytest.mark.usefixtures("setup_class_client")
class TestRelease:
    def test_get_releases(self, datasets, owner) -> None:
        for dataset in datasets:
            dataset_identifier = f"{owner}/{dataset}"
            dataset = self.client.get_dataset(dataset_identifier)
            releases = dataset.get_releases()
            for release in releases:
                assert isinstance(release, Release)

    def test_get_release(self, datasets, owner) -> None:
        for dataset in datasets:
            dataset_identifier = f"{owner}/{dataset}"
            dataset = self.client.get_dataset(dataset_identifier)
            releases = dataset.get_releases()
            for release in releases:
                release = dataset.get_release(release.name)
                assert isinstance(release, Release)

    def test_add_delete_release(self, datasets, owner, TIME_INTERVAL) -> None:
        name = "v0.4"
        description = "Test release description."
        for dataset_name in datasets:
            dataset_identifier = f"{owner}/{dataset_name}"
            dataset = self.client.get_dataset(dataset_identifier)
            if "pointcloud-segmentation" in dataset.task_type:
                # Release files not supported for 3D segmentation
                continue

            release = None
            try:
                # Add release
                release = dataset.add_release(name, description)
                assert isinstance(release, Release)
            except AlreadyExistsError as e:
                release = dataset.get_release(name)
                raise e
            finally:
                # Delete release
                time.sleep(TIME_INTERVAL)
                if release is not None:
                    release.delete()


@pytest.mark.usefixtures("setup_class_client")
class TestLabelset:
    def test_get_labelsets(self, datasets, owner) -> None:
        for dataset in datasets:
            dataset_identifier = f"{owner}/{dataset}"
            dataset = self.client.get_dataset(dataset_identifier)
            labelsets = dataset.get_labelsets()
            for labelset in labelsets:
                assert isinstance(labelset, Labelset)

    def test_get_labelset(self, datasets, owner) -> None:
        for dataset in datasets:
            dataset_identifier = f"{owner}/{dataset}"
            dataset = self.client.get_dataset(dataset_identifier)
            labelsets = dataset.get_labelsets()
            for labelset in labelsets:
                labelset = dataset.get_labelset(labelset.name)
                assert isinstance(labelset, Labelset)

    def test_add_delete_labelset(self, datasets, owner) -> None:
        name = "labelset4"
        description = "Test add_delete_labelset description."
        for dataset in datasets:
            dataset_identifier = f"{owner}/{dataset}"
            dataset = self.client.get_dataset(dataset_identifier)
            labelset = None
            try:
                # Add labelset
                labelset = dataset.add_labelset(name, description)
                assert isinstance(labelset, Labelset)
            except AlreadyExistsError as e:
                labelset = dataset.get_labelset(name)
                raise e
            finally:
                # Delete labelset
                if labelset is not None:
                    labelset.delete()



from typing import List

import pytest
from segments.client import SegmentsClient
from segments.dataset import SegmentsDataset
from segments.utils import export_dataset


def test_export_dataset(
    client: SegmentsClient,
    owner: str,
    datasets: List[str],
    ARTIFACTS_DIR: str,
) -> None:
    # Get the dataset
    dataset_identifier = f"{owner}/{datasets[0]}"
    release = client.get_releases(dataset_identifier)[0]
    dataset = SegmentsDataset(release, segments_dir=f"{ARTIFACTS_DIR}/segments")

    # Export the dataset
    export_formats = [
        "coco-panoptic",
        "coco-instance",
        "yolo",
        "instance",
        "instance-color",
        "semantic",
        "semantic-color",
        "polygon",
    ]
    export_folder = f"{ARTIFACTS_DIR}"
    for export_format in export_formats:
        if export_format == "yolo":
            with pytest.raises(ValueError):
                export_dataset(dataset, export_folder, export_format)
        else:
            export_dataset(dataset, export_folder, export_format)



from __future__ import annotations

from enum import Enum as BaseEnum
from enum import EnumMeta as BaseEnumMeta
from typing import Annotated, Any, Dict, List, Optional, Tuple, Union

import pydantic
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, field_validator
from segments.exceptions import ValidationError
from typing_extensions import Literal, TypedDict


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(
        # What happens with extra fields in dictionaries. Use ignore in production and allow in debug mode https://pydantic-docs.helpmanual.io/usage/model_config/#change-behaviour-globally
        extra="ignore",
        # What happens with wrong field types. Use false in production and true in debug mode https://pydantic-docs.helpmanual.io/usage/types/#arbitrary-types-allowed
        arbitrary_types_allowed=False,
        # Whether to populate models with the value property of enums, rather than the raw enum
        use_enum_values=True,
    )


class EnumMeta(BaseEnumMeta):
    # https://stackoverflow.com/questions/43634618/how-do-i-test-if-int-value-exists-in-python-enum-without-using-try-catch
    def __contains__(self, item):
        return isinstance(item, self) or item in {v.value for v in self.__members__.values()}

    # https://stackoverflow.com/questions/29503339/how-to-get-all-values-from-python-enum-class
    def __str__(self):
        return ", ".join(c.value for c in self)

    def __repr__(self):
        return self.__str__()


class Enum(BaseEnum, metaclass=EnumMeta):
    pass


#################################
# Literals, constants and enums #
#################################
class LabelStatus(str, Enum):
    REVIEWED = "REVIEWED"
    REVIEWING_IN_PROGRESS = "REVIEWING_IN_PROGRESS"
    LABELED = "LABELED"
    LABELING_IN_PROGRESS = "LABELING_IN_PROGRESS"
    REJECTED = "REJECTED"
    PRELABELED = "PRELABELED"
    SKIPPED = "SKIPPED"
    VERIFIED = "VERIFIED"
    UNLABELED = "UNLABELED"


# keep in sync with LabelStatus
class LabelStats(BaseModel):
    REVIEWED: Optional[int] = None
    REVIEWING_IN_PROGRESS: Optional[int] = None
    LABELED: Optional[int] = None
    LABELING_IN_PROGRESS: Optional[int] = None
    REJECTED: Optional[int] = None
    PRELABELED: Optional[int] = None
    SKIPPED: Optional[int] = None
    VERIFIED: Optional[int] = None
    UNLABELED: Optional[int] = None
    # extra
    TOTAL: Optional[int] = None


class TaskType(str, Enum):
    SEGMENTATION_BITMAP = "segmentation-bitmap"
    SEGMENTATION_BITMAP_HIGHRES = "segmentation-bitmap-highres"
    IMAGE_SEGMENTATION_SEQUENCE = "image-segmentation-sequence"
    BBOXES = "bboxes"
    VECTOR = "vector"
    IMAGE_VECTOR_SEQUENCE = "image-vector-sequence"
    KEYPOINTS = "keypoints"
    POINTCLOUD_CUBOID = "pointcloud-cuboid"
    POINTCLOUD_CUBOID_SEQUENCE = "pointcloud-cuboid-sequence"
    POINTCLOUD_SEGMENTATION = "pointcloud-segmentation"
    POINTCLOUD_SEGMENTATION_SEQUENCE = "pointcloud-segmentation-sequence"
    POINTCLOUD_VECTOR = "pointcloud-vector"
    POINTCLOUD_VECTOR_SEQUENCE = "pointcloud-vector-sequence"
    # MULTISENSOR = "multisensor"
    MULTISENSOR_SEQUENCE = (
        "multisensor-sequence"  # combination of pointcloud-cuboid-sequence and image-vector-sequence
    )
    EMPTY = ""


class Role(str, Enum):
    LABELER = "labeler"
    REVIEWER = "reviewer"
    MANAGER = "manager"
    ADMIN = "admin"


class Category(str, Enum):
    STREET_SCENERY = "street_scenery"
    GARDEN = "garden"
    AGRICULTURE = "agriculture"
    SATELLITE = "satellite"
    PEOPLE = "people"
    MEDICAL = "medical"
    FRUIT = "fruit"
    OTHER = "other"


class CameraConvention(str, Enum):
    OPEN_CV = "OpenCV"
    OPEN_GL = "OpenGL"


class InputType(str, Enum):
    SELECT = "select"
    MULTISELECT = "multiselect"
    TEXT = "text"
    NUMBER = "number"
    CHECKBOX = "checkbox"
    VECTOR3 = "vector3"
    QUATERNION = "quaternion"
    POINTS = "points"


class CameraDistortionModel(str, Enum):
    FISH_EYE = "fisheye"
    BROWN_CONRADY = "brown-conrady"


class PCDType(str, Enum):
    PCD = "pcd"
    BINARY_XYZI = "binary-xyzi"
    KITTI = "kitti"
    BINARY_XYZIR = "binary-xyzir"
    NUSCENES = "nuscenes"
    PLY = "ply"
    LAS = "las"
    SPLAT = "splat"


class ReleaseStatus(str, Enum):
    PENDING = "PENDING"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"


class ImageVectorAnnotationType(str, Enum):
    BBOX = "bbox"
    POLYGON = "polygon"
    POLYLINE = "polyline"
    POINT = "point"


class PointcloudCuboidAnnotationType(str, Enum):
    CUBOID = "cuboid"
    CUBOID_SYNC = "cuboid-sync"


PointcloudCuboidAnnotationTypeLiteral = Literal[
    PointcloudCuboidAnnotationType.CUBOID,
    PointcloudCuboidAnnotationType.CUBOID_SYNC,
]


class PointcloudVectorAnnotationType(str, Enum):
    POLYGON = "polygon"
    POLYLINE = "polyline"
    POINT = "point"


PointcloudVectorAnnotationTypeLiteral = Literal[
    PointcloudVectorAnnotationType.POLYGON,
    PointcloudVectorAnnotationType.POLYLINE,
    PointcloudVectorAnnotationType.POINT,
]


class ExportFormat(str, Enum):
    COCO_PANOPTIC = "coco-panoptic"
    COCO_INSTANCE = "coco-instance"
    YOLO = "yolo"
    INSTANCE = "instance"
    INSTANCE_COLOR = "instance-color"
    SEMANTIC = "semantic"
    SEMANTIC_COLOR = "semantic-color"
    POLYGON = "polygon"


class Subscription(str, Enum):
    FREE = "FREE"
    STANDARD = "STANDARD"
    ENTERPRISE = "ENTERPRISE"
    ACADEMIC = "ACADEMIC"
    TRIAL = "TRIAL"


RGB = Tuple[int, int, int]
RGBA = Tuple[int, int, int, int]
FormatVersion = Union[float, str]
ObjectAttributes = Dict[str, Optional[Union[str, bool, int, float]]]
ImageAttributes = Dict[str, Optional[Union[str, bool, int, float]]]
LinkAttributes = Dict[str, Optional[Union[str, bool, int, float]]]
Timestamp = Union[str, int, float]


###########
# Release #
###########
class URL(BaseModel):
    # TODO Remove optional (the backend does not return an URL when adding a release)
    url: Optional[str] = None
    signed_url: Optional[str] = None


class DeleteSamplesResponse(BaseModel):
    """Response from bulk sample deletion.

    Attributes:
        deleted_count: Number of samples successfully deleted.
    """

    deleted_count: int


class Release(BaseModel):
    uuid: str
    name: str
    description: str
    release_type: Literal["JSON"]
    attributes: URL
    status: ReleaseStatus
    # status_info: str
    created_at: str
    samples_count: int


#########
# Issue #
#########


# Anchor coordinates
class IssueAnchor2DCoordinates(BaseModel):
    type: Literal["2d"]
    x: float
    y: float


class IssueAnchor3DCoordinates(BaseModel):
    type: Literal["3d"]
    x: float
    y: float
    z: float


IssueAnchorCoordinates = Annotated[
    Union[IssueAnchor2DCoordinates, IssueAnchor3DCoordinates],
    pydantic.Field(discriminator="type"),
]
"""IssueAnchorCoordinates can be one of the following:

- :py:class:`IssueAnchor2DCoordinates`
- :py:class:`IssueAnchor3DCoordinates`
"""


# Anchors
class IssueBaseAnchor(BaseModel):
    sensor: Optional[str] = None
    frame: Optional[int] = None
    type: Literal["basic", "location", "sceneAttribute", "location"]


class IssueBasicAnchor(IssueBaseAnchor):
    type: Literal["basic"]


class IssueTrackAnchor(IssueBaseAnchor):
    type: Literal["track"]
    track: int
    trackAttribute: Optional[str] = None


class IssueSceneAttributeAnchor(IssueBaseAnchor):
    type: Literal["sceneAttribute"]
    sceneAttribute: str


class IssueLocationAnchor(IssueBaseAnchor):
    type: Literal["location"]
    location: IssueAnchorCoordinates


IssueAnchor = Annotated[
    Union[
        IssueBasicAnchor,
        IssueTrackAnchor,
        IssueSceneAttributeAnchor,
        IssueLocationAnchor,
    ],
    pydantic.Field(discriminator="type"),
]
"""IssueAnchor can be one of the following:

- :py:class:`IssueBasicAnchor`
- :py:class:`IssueTrackAnchor`
- :py:class:`IssueSceneAttributeAnchor`
- :py:class:`IssueLocationAnchor`
"""


# Issue
class IssueStatus(str, Enum):
    OPEN = "OPEN"
    CLOSED = "CLOSED"


class IssueComment(BaseModel):
    created_at: str
    created_by: str
    text: str


class Issue(BaseModel):
    uuid: str
    description: str
    created_at: str
    updated_at: str
    created_by: str
    updated_by: str
    status: IssueStatus
    comments: List[IssueComment]
    anchor: Optional[IssueAnchor] = None
    sample_uuid: str
    sample_name: str


########
# File #
########
# https://stackoverflow.com/questions/60003444/typeddict-when-keys-have-invalid-names
AWSFields = TypedDict(
    "AWSFields",
    {
        "acl": str,
        "Content-Type": str,
        "key": str,
        "x-amz-algorithm": str,
        "x-amz-credential": str,
        "x-amz-date": str,
        "policy": str,
        "x-amz-signature": str,
    },
)


class PresignedPostFields(BaseModel):
    url: str
    fields: AWSFields


class File(BaseModel):
    uuid: str
    filename: str
    url: str
    created_at: str
    presignedPostFields: PresignedPostFields


#########
# Label #
#########
class Annotation(BaseModel):
    id: int
    category_id: int
    attributes: Optional[ObjectAttributes] = None


class Link(BaseModel):
    from_id: int
    to_id: int
    attributes: Optional[LinkAttributes] = None


# Image segmentation
class ImageSegmentationLabelAttributes(BaseModel):
    annotations: List[Annotation]
    segmentation_bitmap: URL
    image_attributes: Optional[ImageAttributes] = None
    format_version: Optional[FormatVersion] = None


# Image vector
# https://stackoverflow.com/questions/51575931/class-inheritance-in-python-3-7-dataclasses
class ImageVectorAnnotation(BaseModel):
    id: int
    category_id: int
    points: List[List[float]]
    type: ImageVectorAnnotationType
    attributes: Optional[ObjectAttributes] = None


class ImageVectorLabelAttributes(BaseModel):
    annotations: List[ImageVectorAnnotation]
    format_version: Optional[FormatVersion] = None
    image_attributes: Optional[ImageAttributes] = None


# Image sequence segmentation
class ImageSequenceSegmentationAnnotation(Annotation):
    track_id: int
    is_keyframe: bool = False


class ImageSequenceSegmentationFrame(ImageSegmentationLabelAttributes):
    annotations: List[ImageSequenceSegmentationAnnotation]
    timestamp: Optional[Timestamp] = None
    format_version: Optional[FormatVersion] = None


class ImageSequenceSegmentationLabelAttributes(BaseModel):
    frames: List[ImageSequenceSegmentationFrame]
    format_version: Optional[FormatVersion] = None


# Image sequence vector
class ImageSequenceVectorAnnotation(ImageVectorAnnotation):
    track_id: int
    is_keyframe: bool = False


class ImageVectorFrame(ImageVectorLabelAttributes):
    annotations: List[ImageSequenceVectorAnnotation]
    timestamp: Optional[Timestamp] = None
    format_version: Optional[FormatVersion] = None
    image_attributes: Optional[ImageAttributes] = None


class ImageSequenceVectorLabelAttributes(BaseModel):
    frames: List[ImageVectorFrame]
    format_version: Optional[FormatVersion] = None


# Point cloud segmentation
class PointcloudSegmentationLabelAttributes(BaseModel):
    annotations: List[Annotation]
    image_attributes: Optional[ImageAttributes] = None
    point_annotations: List[int]
    format_version: Optional[FormatVersion] = None


class XYZ(BaseModel):
    x: float
    y: float
    z: float


class PartialXYZ(BaseModel):
    x: Optional[float] = None
    y: Optional[float] = None
    z: Optional[float] = None


class XYZW(BaseModel):
    qx: float
    qy: float
    qz: float
    qw: float


class Velocity(BaseModel):
    relative: Optional[float]
    absolute: float


class Acceleration(BaseModel):
    relative: Optional[float]
    absolute: float


class FisheyeDistortionCoefficients(BaseModel):
    k1: float
    k2: float
    k3: float
    k4: float


class BrownConradyDistortionCoefficients(BaseModel):
    k1: float
    k2: float
    p1: float
    p2: float
    k3: float
    k4: Optional[float] = None
    k5: Optional[float] = None
    k6: Optional[float] = None


class DistortionFisheye(BaseModel):
    model: Literal[CameraDistortionModel.FISH_EYE]
    coefficients: FisheyeDistortionCoefficients


class DistortionBrownConrady(BaseModel):
    model: Literal[CameraDistortionModel.BROWN_CONRADY]
    coefficients: BrownConradyDistortionCoefficients


Distortion = Annotated[Union[DistortionFisheye, DistortionBrownConrady], pydantic.Field(discriminator="model")]


# Point cloud cuboid
# https://stackoverflow.com/questions/51575931/class-inheritance-in-python-3-7-dataclasses
class PointcloudCuboidAnnotation(BaseModel):
    id: int
    category_id: int
    position: XYZ
    dimensions: XYZ
    yaw: float
    rotation: Optional[XYZW] = None
    type: PointcloudCuboidAnnotationTypeLiteral
    attributes: Optional[ObjectAttributes] = None
    velocity: Optional[Velocity] = None
    acceleration: Optional[Acceleration] = None


class BasePointcloudCuboidLabelAttributes(BaseModel):
    annotations: List[PointcloudCuboidAnnotation]
    image_attributes: Optional[ImageAttributes] = None
    format_version: Optional[FormatVersion] = None


class PointcloudCuboidLabelAttributes(BasePointcloudCuboidLabelAttributes):
    links: Optional[List[Link]] = None


# Point cloud vector
class PointcloudVectorAnnotation(BaseModel):
    id: int
    category_id: int
    points: List[List[float]]
    type: PointcloudVectorAnnotationTypeLiteral
    attributes: Optional[ObjectAttributes] = None


class BasePointcloudVectorLabelAttributes(BaseModel):
    annotations: List[
        Annotated[Union[PointcloudVectorAnnotation, PointcloudCuboidAnnotation], pydantic.Field(discriminator="type")]
    ]
    image_attributes: Optional[ImageAttributes] = None
    format_version: Optional[FormatVersion] = None


class PointcloudVectorLabelAttributes(BasePointcloudVectorLabelAttributes):
    links: Optional[List[Link]] = None


# Point cloud sequence segmentation
class PointcloudSequenceSegmentationAnnotation(Annotation):
    track_id: int
    is_keyframe: bool = False
    attributes: Optional[ObjectAttributes] = None


class PointcloudSegmentationFrame(PointcloudSegmentationLabelAttributes):
    annotations: List[PointcloudSequenceSegmentationAnnotation]
    timestamp: Optional[Timestamp] = None


class PointcloudSequenceSegmentationLabelAttributes(BaseModel):
    frames: List[PointcloudSegmentationFrame]
    format_version: Optional[FormatVersion] = None


# Point cloud sequence cuboid
class PointcloudSequenceCuboidAnnotation(PointcloudCuboidAnnotation):
    track_id: int
    is_keyframe: bool = False


class PointcloudSequenceCuboidFrame(BasePointcloudCuboidLabelAttributes):
    annotations: List[PointcloudSequenceCuboidAnnotation]
    timestamp: Optional[Timestamp] = None
    format_version: Optional[FormatVersion] = None


class BasePointcloudSequenceCuboidLabelAttributes(BaseModel):
    frames: List[PointcloudSequenceCuboidFrame]
    format_version: Optional[FormatVersion] = None


class PointcloudSequenceCuboidLabelAttributes(BasePointcloudSequenceCuboidLabelAttributes):
    links: Optional[List[Link]] = None


# Point cloud sequence vector
class PointcloudSequenceVectorAnnotation(PointcloudVectorAnnotation):
    track_id: int
    is_keyframe: bool = False


class PointcloudSequenceVectorFrame(PointcloudVectorLabelAttributes):
    annotations: List[
        Annotated[
            Union[PointcloudSequenceVectorAnnotation, PointcloudSequenceCuboidAnnotation],
            pydantic.Field(discriminator="type"),
        ]
    ]
    timestamp: Optional[Timestamp] = None


class BasePointcloudSequenceVectorLabelAttributes(BaseModel):
    frames: List[PointcloudSequenceVectorFrame]
    format_version: Optional[FormatVersion] = None


class PointcloudSequenceVectorLabelAttributes(BasePointcloudSequenceVectorLabelAttributes):
    links: Optional[List[Link]] = None


# Multi-sensor
class MultiSensorPointcloudSequenceCuboidLabelAttributes(BaseModel):
    name: str
    task_type: Literal[TaskType.POINTCLOUD_CUBOID_SEQUENCE]
    # TODO remove list and replace with `Optional[PointcloudSequenceCuboidLabelAttributes] = None`
    attributes: Union[BasePointcloudSequenceCuboidLabelAttributes, List]


class MultiSensorImageSequenceVectorLabelAttributes(BaseModel):
    name: str
    task_type: Literal[TaskType.IMAGE_VECTOR_SEQUENCE]
    # TODO remove list and replace with `Optional[ImageSequenceVectorLabelAttributes] = None`
    attributes: Union[ImageSequenceVectorLabelAttributes, List]


class MultiSensorPointcloudSequenceVectorLabelAttributes(BaseModel):
    name: str
    task_type: Literal[TaskType.POINTCLOUD_VECTOR_SEQUENCE]
    # TODO remove list and replace with `Optional[ImageSequenceVectorLabelAttributes] = None`
    attributes: Union[BasePointcloudSequenceVectorLabelAttributes, List]


class MultiSensorLabelAttributes(BaseModel):
    sensors: List[
        Annotated[
            Union[
                MultiSensorPointcloudSequenceCuboidLabelAttributes,
                MultiSensorPointcloudSequenceVectorLabelAttributes,
                MultiSensorImageSequenceVectorLabelAttributes,
            ],
            pydantic.Field(discriminator="task_type"),
        ]
    ]
    links: Optional[List[Link]] = None


LabelAttributes = Union[
    ImageVectorLabelAttributes,
    ImageSegmentationLabelAttributes,
    ImageSequenceVectorLabelAttributes,
    ImageSequenceSegmentationLabelAttributes,
    PointcloudCuboidLabelAttributes,
    PointcloudVectorLabelAttributes,
    PointcloudSegmentationLabelAttributes,
    PointcloudSequenceCuboidLabelAttributes,
    PointcloudSequenceVectorLabelAttributes,
    PointcloudSequenceSegmentationLabelAttributes,
    MultiSensorLabelAttributes,
]


class Label(BaseModel):
    sample_uuid: str
    label_type: TaskType
    label_status: LabelStatus
    labelset: str
    attributes: Optional[LabelAttributes] = None
    created_at: str
    created_by: str
    updated_at: str
    updated_by: Optional[str] = None
    score: Optional[float] = None
    rating: Optional[float] = None
    reviewed_at: Optional[str] = None
    reviewed_by: Optional[str] = None


class LabelSummary(BaseModel):
    score: Optional[float] = None
    label_status: LabelStatus
    updated_at: Optional[str] = None
    created_by: Optional[str] = None
    reviewed_by: Optional[str] = None


##########
# Sample #
##########
# Image
class ImageSampleAttributes(BaseModel):
    image: URL


# Image sequence
class ImageFrame(ImageSampleAttributes):
    name: Optional[str] = None
    timestamp: Optional[Timestamp] = None


class ImageSequenceSampleAttributes(BaseModel):
    frames: List[ImageFrame]


# Point cloud
class PCD(BaseModel):
    url: str
    signed_url: Optional[str] = None
    type: PCDType


class EgoPose(BaseModel):
    position: XYZ
    heading: XYZW


class Bounds(BaseModel):
    min_x: Optional[float] = None
    max_x: Optional[float] = None
    min_y: Optional[float] = None
    max_y: Optional[float] = None
    min_z: Optional[float] = None
    max_z: Optional[float] = None


class CameraIntrinsics(BaseModel):
    intrinsic_matrix: List[List[float]]


class CameraExtrinsics(BaseModel):
    translation: XYZ
    rotation: XYZW


class CalibratedImage(URL):
    row: Optional[int] = None
    col: Optional[int] = None
    intrinsics: Optional[CameraIntrinsics] = None
    extrinsics: Optional[CameraExtrinsics] = None
    distortion: Optional[Distortion] = None
    camera_convention: CameraConvention = CameraConvention.OPEN_GL
    name: Optional[str] = None
    rotation: Optional[float] = None


class PointcloudSampleAttributes(BaseModel):
    pcd: PCD
    images: List[CalibratedImage] = []
    ego_pose: Optional[EgoPose] = None
    default_z: Optional[float] = None
    name: Optional[str] = None
    timestamp: Optional[Timestamp] = None
    bounds: Optional[Bounds] = None


# Point cloud sequence
class PointcloudSequenceSampleAttributes(BaseModel):
    frames: List[PointcloudSampleAttributes]


# Multi-sensor
class MultiSensorPointcloudSequenceSampleAttributes(BaseModel):
    name: str
    task_type: Union[
        Literal[TaskType.POINTCLOUD_CUBOID_SEQUENCE],
        Literal[TaskType.POINTCLOUD_VECTOR_SEQUENCE],
    ]
    attributes: PointcloudSequenceSampleAttributes


class MultiSensorImageSequenceSampleAttributes(BaseModel):
    name: str
    task_type: Literal[TaskType.IMAGE_VECTOR_SEQUENCE]
    attributes: ImageSequenceSampleAttributes


class MultiSensorSampleAttributes(BaseModel):
    sensors: List[
        Union[
            MultiSensorPointcloudSequenceSampleAttributes,
            MultiSensorImageSequenceSampleAttributes,
        ],
    ]


SampleAttributes = Union[
    ImageSampleAttributes,
    ImageSequenceSampleAttributes,
    PointcloudSampleAttributes,
    PointcloudSequenceSampleAttributes,
    MultiSensorSampleAttributes,
]


class Sample(BaseModel):
    uuid: str
    name: str
    attributes: Optional[SampleAttributes] = None
    metadata: Dict[str, Any]
    created_at: str
    created_by: str
    readme: str = ""
    assigned_labeler: Optional[str] = None
    assigned_reviewer: Optional[str] = None
    priority: float
    label: Optional[Union[Label, LabelSummary]] = None
    issues: Optional[List[Issue]] = None
    dataset_full_name: Optional[str] = None


########################
# Dataset and labelset #
########################
# https://docs.pydantic.dev/latest/concepts/postponed_annotations/#self-referencing-or-recursive-models
class User(BaseModel):
    username: str
    created_at: str
    is_organization: bool
    email: Optional[str] = None
    webhooks_enabled: Optional[bool] = None
    private_upload_count: Optional[int] = None
    public_upload_count: Optional[int] = None
    subscription: Optional[Subscription] = None
    is_trial_expired: Optional[bool] = None
    organizations: Optional[List[User]] = None
    organization_created_by: Optional[str] = None
    organization_role: Optional[Role] = None
    members: Optional[List[User]] = None
    insights_urls: Optional[Dict[str, str]] = None


class Collaborator(BaseModel):
    user: User
    role: Role


class BaseAttribute(BaseModel):
    name: str
    is_mandatory: Optional[bool] = None


class BaseSelectTaskAttribute(BaseAttribute):
    input_type: Literal[InputType.SELECT]
    values: List[str]
    default_value: Optional[str] = None


class BaseMultiselectTaskAttribute(BaseAttribute):
    input_type: Literal[InputType.MULTISELECT]
    values: List[str]
    default_value: Optional[str] = None


class BaseTextTaskAttribute(BaseAttribute):
    input_type: Literal[InputType.TEXT]
    default_value: Optional[str] = None


class BaseNumberTaskAttribute(BaseAttribute):
    input_type: Literal[InputType.NUMBER]
    default_value: Optional[float] = None
    min: Optional[float] = None
    max: Optional[float] = None
    step: Optional[float] = None

    @field_validator("min", "max", "step", mode="before")
    @classmethod
    def empty_str_to_none(cls, v):
        # `min`, `max` and `step` are empty strings when not filled in
        if isinstance(v, str) and v.strip() == "":
            return None
        return v


class BaseCheckboxTaskAttribute(BaseAttribute):
    input_type: Literal[InputType.CHECKBOX]
    default_value: Optional[bool] = None


class BaseVector3TaskAttribute(BaseAttribute):
    input_type: Literal[InputType.VECTOR3]


class BaseQuaternionTaskAttribute(BaseAttribute):
    input_type: Literal[InputType.QUATERNION]


class BasePointsTaskAttribute(BaseAttribute):
    input_type: Literal[InputType.POINTS]


class BaseTaskAttribute(BaseModel):
    is_track_level: Optional[bool] = None
    synced_across_sensors: Optional[bool] = None
    sensor_filter: Optional[str] = None


class SelectTaskAttribute(BaseTaskAttribute, BaseSelectTaskAttribute):
    pass


class MultiselectTaskAttribute(BaseTaskAttribute, BaseMultiselectTaskAttribute):
    pass


class TextTaskAttribute(BaseTaskAttribute, BaseTextTaskAttribute):
    pass


class NumberTaskAttribute(BaseTaskAttribute, BaseNumberTaskAttribute):
    pass


class CheckboxTaskAttribute(BaseTaskAttribute, BaseCheckboxTaskAttribute):
    pass


class Vector3TaskAttribute(BaseTaskAttribute, BaseVector3TaskAttribute):
    pass


class QuaternionTaskAttribute(BaseTaskAttribute, BaseQuaternionTaskAttribute):
    pass


class PointsTaskAttribute(BaseTaskAttribute, BasePointsTaskAttribute):
    pass


TaskAttribute = Annotated[
    Union[
        SelectTaskAttribute,
        MultiselectTaskAttribute,
        TextTaskAttribute,
        NumberTaskAttribute,
        CheckboxTaskAttribute,
        Vector3TaskAttribute,
        QuaternionTaskAttribute,
        PointsTaskAttribute,
    ],
    pydantic.Field(discriminator="input_type"),
]
"""TaskAttribute can be one of the following:

- :py:class:`SelectTaskAttribute`
- :py:class:`MultiselectTaskAttribute`
- :py:class:`TextTaskAttribute`
- :py:class:`NumberTaskAttribute`
- :py:class:`CheckboxTaskAttribute`
- :py:class:`Vector3TaskAttribute`
- :py:class:`QuaternionTaskAttribute`
- :py:class:`PointsTaskAttribute`
"""


class BaseTaskLinkAttribute(BaseModel):
    is_track_level: Literal[True] = True


class SelectTaskLinkAttribute(BaseTaskLinkAttribute, BaseSelectTaskAttribute):
    pass


class MultiselectTaskLinkAttribute(BaseTaskLinkAttribute, BaseMultiselectTaskAttribute):
    pass


class TextTaskLinkAttribute(BaseTaskLinkAttribute, BaseTextTaskAttribute):
    pass


class NumberTaskLinkAttribute(BaseTaskLinkAttribute, BaseNumberTaskAttribute):
    pass


class CheckboxTaskLinkAttribute(BaseTaskLinkAttribute, BaseCheckboxTaskAttribute):
    pass


TaskLinkAttribute = Annotated[
    Union[
        SelectTaskLinkAttribute,
        MultiselectTaskLinkAttribute,
        TextTaskLinkAttribute,
        NumberTaskLinkAttribute,
        CheckboxTaskLinkAttribute,
    ],
    pydantic.Field(discriminator="input_type"),
]
"""TaskLinkAttribute can be one of the following:


- :py:class:`SelectTaskLinkAttribute`
- :py:class:`MultiselectTaskLinkAttribute`
- :py:class:`TextTaskLinkAttribute`
- :py:class:`NumberTaskLinkAttribute`
- :py:class:`CheckboxTaskLinkAttribute`
"""


class TaskAttributeCategory(BaseModel):
    name: str
    id: int
    color: Optional[Union[RGB, RGBA]] = None
    has_instances: Optional[bool] = None
    attributes: Optional[List[TaskAttribute]] = None
    dimensions: Optional[XYZ] = None
    model_config = ConfigDict(extra="allow")
    link_attributes: Optional[List[TaskLinkAttribute]] = None


class IntersectingCuboidsRule(BaseModel):
    name: Literal["intersecting-cuboids"]
    excluded_categories: Optional[List[int]] = None
    excluded_set_of_categories: Optional[List[int]] = None


class CuboidDimensionLimitsRule(BaseModel):
    name: Literal["cuboid-dimension-limits"]
    categories: Optional[List[int]] = None
    min: Optional[PartialXYZ] = None
    max: Optional[PartialXYZ] = None


WarningRule = Annotated[
    Union[
        IntersectingCuboidsRule,
        CuboidDimensionLimitsRule,
    ],
    pydantic.Field(discriminator="name"),
]
"""WarningRule can be one of the following:

- :py:class:`IntersectingCuboidsRule`
- :py:class:`CuboidDimensionLimitsRule`
"""


class TaskAttributes(BaseModel):
    format_version: Optional[FormatVersion] = None
    categories: Optional[List[TaskAttributeCategory]] = None
    image_attributes: Optional[List[TaskAttribute]] = None
    model_config = ConfigDict(extra="allow")
    warning_rules: Optional[List[WarningRule]] = None


class Owner(BaseModel):
    username: str
    created_at: str
    email: Optional[str] = None


class Labelset(BaseModel):
    name: str
    description: str
    # uuid: Optional[str] = None
    # readme: Optional[str] = None
    # task_type: Optional[TaskType] = None
    # attributes: Optional[Union[str, TaskAttributes]] = None
    is_groundtruth: Optional[bool] = None
    # statistics: Optional[Statistics] = None
    created_at: Optional[str] = None
    # stats: Optional[Dict[str, Any]] = None


class Workunit(BaseModel):
    uuid: Optional[str] = None
    created_at: str
    created_by: str
    work_type: str
    next_label_status: Optional[str] = None
    time: int
    inactive_time: int
    sample_uuid: str
    sample_name: str
    label_rating: Optional[int] = None


class Dataset(BaseModel):
    name: str
    full_name: str
    cloned_from: Optional[str] = None
    description: str
    # data_type: Literal["IMAGE"]
    category: Category
    public: bool
    archived: bool
    owner: Owner
    created_at: str
    enable_ratings: bool
    enable_skip_labeling: bool
    enable_skip_reviewing: bool
    enable_save_button: bool
    enable_label_status_verified: bool
    enable_same_dimensions_track_constraint: bool
    enable_3d_region_of_interest: bool
    enable_interpolation: bool
    use_timestamps_for_interpolation: bool
    enable_object_linking: bool
    task_type: TaskType
    # task_readme: str
    label_stats: LabelStats
    labeling_inactivity_timeout_seconds: Optional[int] = None
    samples_count: Optional[Union[str, int]] = None
    collaborators_count: Optional[int] = None
    task_attributes: Optional[TaskAttributes] = None
    labelsets: Optional[List[Labelset]] = None
    role: Optional[str] = None
    readme: Optional[str] = None
    metadata: Dict[str, Any]
    noncollaborator_can_label: Optional[bool] = None
    noncollaborator_can_review: Optional[bool] = None
    insights_urls: Optional[Dict[str, str]] = None
    region_of_interest: Optional[Dict[str, Any]] = None
    # tasks: Optional[List[Dict[str, Any]]] = None

    @field_validator("category")
    @classmethod
    def check_category(cls, category: str) -> str:
        if category not in Category and "custom-" not in category:
            raise ValidationError(f"The category should be one of {Category}, but is {category}.")
        return category


####################
# Segments dataset #
####################
class SegmentsDatasetCategory(BaseModel):
    id: int
    name: str
    color: Optional[Union[RGB, RGBA]] = None
    attributes: Optional[List[Any]] = None
    model_config = ConfigDict(populate_by_name=True, extra="allow")



from typing import Optional


class SegmentsError(Exception):
    """Base class for exceptions."""

    def __init__(self, message: str, cause: Optional[Exception] = None):
        """
        Args:
            message: An informative message about the exception.
            cause: The cause of the exception raised by Python or another library. Defaults to :obj:`None`.
        """

        super().__init__(message, cause)
        self.message = message
        self.cause = cause

    def __str__(self) -> str:
        return self.message


##################
# Authentication #
##################
class AuthenticationError(SegmentsError):
    """Raised when an API key fails authentication."""

    pass


#################
# Authorization #
#################
class AuthorizationError(SegmentsError):
    """Raised when a user is unauthorized to perform the given request."""

    pass


########################
# Network (HTTP error) #
########################
class NetworkError(SegmentsError):
    """Raised when an HTTP error occurs."""

    pass


class NotFoundError(NetworkError):
    """Raised when the requested object is not found (e.g., because the name is misspelled, the page index is incorrect, ...)."""

    pass


class AlreadyExistsError(NetworkError):
    """Raised when the object (e.g., dataset, labelset, release, ...) already exists."""

    pass


class CollaboratorError(NetworkError):
    """Raised when a collaborator error occurs (e.g., a collaborator already has a role / cannot be added to an organization / ...)."""

    pass


class SubscriptionError(NetworkError):
    """Raised when a subscription error occurs (e.g., your trial ends, you exceed your user limit for private datasets, ...)."""

    pass


############
# Time out #
############
class TimeoutError(SegmentsError):
    """Raised when a request times out."""

    pass


#############
# API limit #
#############
class APILimitError(SegmentsError):
    """Raised when the user performs too many requests in a period of time."""

    pass


##############
# Validation #
##############
class ValidationError(SegmentsError):
    """Raised when validation of the response fails."""

    pass


class InvalidModelError(SegmentsError):
    """Raised when the model is invalid for the dataset type."""

    pass


#############
# SDK Usage #
#############
class MissingContextError(SegmentsError):
    """Raised when the SDK is missing context information about a resource."""

    pass



from __future__ import annotations

import copy
import json
import logging
import os
import random
import re
from collections import defaultdict
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Tuple, Union, cast
from urllib.parse import urlparse

import numpy as np
import numpy.typing as npt
import requests
from PIL import ExifTags, Image
from segments import SegmentsClient
from segments.exceptions import AlreadyExistsError
from segments.typing import (
    XYZW,
    Acceleration,
    Annotation,
    EgoPose,
    ExportFormat,
    Label,
    MultiSensorLabelAttributes,
    MultiSensorPointcloudSequenceCuboidLabelAttributes,
    MultiSensorSampleAttributes,
    PointcloudCuboidLabelAttributes,
    PointcloudSequenceCuboidLabelAttributes,
    PointcloudSequenceSampleAttributes,
    Sample,
    TaskType,
    Velocity,
)
from tqdm import tqdm


# https://adamj.eu/tech/2021/05/13/python-type-hints-how-to-fix-circular-imports/
# https://stackoverflow.com/questions/61384752/how-to-type-hint-with-an-optional-import
if TYPE_CHECKING:
    import open3d as o3d
    from pyquaternion import Quaternion
    from segments.dataset import SegmentsDataset
    from segments.typing import Release


#############
# Variables #
#############
session = requests.Session()
adapter = requests.adapters.HTTPAdapter(max_retries=3)
session.mount("http://", adapter)
session.mount("https://", adapter)
logger = logging.getLogger(__name__)
SEGMENTS_ORANGE = "#FF9900"
COMPATIBLE_TASK_TYPES = {
    ExportFormat.COCO_PANOPTIC: {
        TaskType.SEGMENTATION_BITMAP,
        TaskType.SEGMENTATION_BITMAP_HIGHRES,
    },
    ExportFormat.COCO_INSTANCE: {
        TaskType.VECTOR,
        TaskType.BBOXES,
        TaskType.SEGMENTATION_BITMAP,
        TaskType.SEGMENTATION_BITMAP_HIGHRES,
    },
    ExportFormat.YOLO: {
        TaskType.VECTOR,
        TaskType.BBOXES,
        TaskType.IMAGE_VECTOR_SEQUENCE,
    },
    ExportFormat.INSTANCE: {
        TaskType.SEGMENTATION_BITMAP,
        TaskType.SEGMENTATION_BITMAP_HIGHRES,
    },
    ExportFormat.INSTANCE_COLOR: {
        TaskType.SEGMENTATION_BITMAP,
        TaskType.SEGMENTATION_BITMAP_HIGHRES,
    },
    ExportFormat.SEMANTIC: {
        TaskType.SEGMENTATION_BITMAP,
        TaskType.SEGMENTATION_BITMAP_HIGHRES,
    },
    ExportFormat.SEMANTIC_COLOR: {
        TaskType.SEGMENTATION_BITMAP,
        TaskType.SEGMENTATION_BITMAP_HIGHRES,
    },
    ExportFormat.POLYGON: {
        TaskType.SEGMENTATION_BITMAP,
        TaskType.SEGMENTATION_BITMAP_HIGHRES,
    },
}


def bitmap2file(
    bitmap: npt.NDArray[np.uint32],
    is_segmentation_bitmap: bool = True,
) -> BytesIO:
    """Convert a label bitmap to a file with the proper format.

    Args:
        bitmap: A :class:`numpy.ndarray` with :class:`numpy.uint32` dtype where each unique value represents an instance id.
        is_segmentation_bitmap: If this is a segmentation bitmap. Defaults to :obj:`True`.

    Returns:
        A file object.

    Raises:
        :exc:`ValueError`: If the ``dtype`` is not :class:`np.uint32` or :class:`np.uint8`.
        :exc:`ValueError`: If the bitmap is not a segmentation bitmap.
    """

    # Convert bitmap to np.uint32, if it is not already
    if bitmap.dtype == "uint32":
        pass
    elif bitmap.dtype == "uint8":
        bitmap = np.uint32(bitmap)
    else:
        raise ValueError("Only `np.ndarrays` with `np.uint32` data type can be used.")

    if is_segmentation_bitmap:
        bitmap2 = np.copy(bitmap)
        bitmap2 = bitmap2[:, :, None].view(np.uint8)
        bitmap2[:, :, 3] = 255
    else:
        raise ValueError("Only segmentation bitmaps can be used.")

    f = BytesIO()
    Image.fromarray(bitmap2).save(f, "PNG")
    f.seek(0)
    return f


def get_semantic_bitmap(
    instance_bitmap: Optional[npt.NDArray[np.uint32]] = None,
    annotations: Optional[Union[List[Dict[str, Any], List[Annotation]]]] = None,
    id_increment: int = 0,
) -> Optional[npt.NDArray[np.uint32]]:
    """Convert an instance bitmap and annotations dict into a segmentation bitmap.

    Args:
        instance_bitmap: A :class:`numpy.ndarray` with :class:`numpy.uint32` ``dtype`` where each unique value represents an instance id. Defaults to :obj:`None`.
        annotations: An annotations dictionary. Defaults to :obj:`None`.
        id_increment: Increment the category ids with this number. Defaults to ``0``.

    Returns:
        An array here each unique value represents a category id.
    """
    # Check if annotations isa list of Annotation objects, if they are, convert them.
    if isinstance(annotations[0], Annotation):
        annotations = [annotation.model_dump() for annotation in annotations]

    if instance_bitmap is None or annotations is None:
        return None

    instance2semantic = [0] * (max([a["id"] for a in annotations], default=0) + 1)
    for annotation in annotations:
        instance2semantic[annotation["id"]] = annotation["category_id"] + id_increment
    instance2semantic = np.array(instance2semantic)

    semantic_label = instance2semantic[np.array(instance_bitmap, np.uint32)]
    return semantic_label


def export_dataset(
    dataset: SegmentsDataset,
    export_folder: str = ".",
    export_format: ExportFormat = ExportFormat.COCO_PANOPTIC,
    id_increment: int = 0,
    **kwargs: Any,
) -> Optional[Union[Tuple[str, Optional[str]], Optional[str]]]:
    """Export a dataset to a different format.

    +------------------+-------------------------------------------------------------+
    | Export format    | Supported dataset type                                      |
    +==================+=============================================================+
    | COCO panoptic    | ``segmentation-bitmap`` and ``segmentation-bitmap-highres`` |
    +------------------+-------------------------------------------------------------+
    | COCO instance    | ``segmentation-bitmap`` and ``segmentation-bitmap-highres`` |
    +------------------+-------------------------------------------------------------+
    | YOLO             | ``vector``, ``bboxes`` and ``image-vector-sequence``        |
    +------------------+-------------------------------------------------------------+
    | Instance         | ``segmentation-bitmap`` and ``segmentation-bitmap-highres`` |
    +------------------+-------------------------------------------------------------+
    | Colored instance | ``segmentation-bitmap`` and ``segmentation-bitmap-highres`` |
    +------------------+-------------------------------------------------------------+
    | Semantic         | ``segmentation-bitmap`` and ``segmentation-bitmap-highres`` |
    +------------------+-------------------------------------------------------------+
    | Colored semantic | ``segmentation-bitmap`` and ``segmentation-bitmap-highres`` |
    +------------------+-------------------------------------------------------------+
    | Polygon          | ``segmentation-bitmap`` and ``segmentation-bitmap-highres`` |
    +------------------+-------------------------------------------------------------+

    Example:

    .. code-block:: python

        # pip install segments-ai
        from segments import SegmentsClient, SegmentsDataset
        from segments.utils import export_dataset

        # Initialize a SegmentsDataset from the release file
        client = SegmentsClient('YOUR_API_KEY')
        release = client.get_release('jane/flowers', 'v1.0') # Alternatively: release = 'flowers-v1.0.json'
        dataset = SegmentsDataset(release, labelset='ground-truth', filter_by=['labeled', 'reviewed'])

        # Export to COCO panoptic format
        export_dataset(dataset, export_format='coco-panoptic')


    Alternatively, you can use the initialized :class:`.SegmentsDataset` to loop through the samples and labels, and visualize or process them in any way you please:

    .. code-block:: python

        import matplotlib.pyplot as plt
        from segments.utils import get_semantic_bitmap

        for sample in dataset:
            # Print the sample name and list of labeled objects
            print(sample['name'])
            print(sample['annotations'])

            # Show the image
            plt.imshow(sample['image'])
            plt.show()

            # Show the instance segmentation label
            plt.imshow(sample['segmentation_bitmap'])
            plt.show()

            # Show the semantic segmentation label
            semantic_bitmap = get_semantic_bitmap(sample['segmentation_bitmap'], sample['annotations'])
            plt.imshow(semantic_bitmap)
            plt.show()

    Args:
        dataset: A :class:`.SegmentsDataset`.
        export_folder: The folder to export the dataset to. Defaults to ``.``.
        export_format: The destination format. Defaults to ``coco-panoptic``.
        id_increment: Increment the category ids with this number. Defaults to ``0``. Ignored unless ``export_format`` is ``semantic`` or ``semantic-color``.

    Returns:
        Returns the file name and the image directory name (for COCO panoptic, COCO instance, YOLO and polygon), or returns the export folder name (for (colored) instance and (colored) panoptic).

    Raises:
        :exc:`ImportError`: If scikit image is not installed (to install run ``pip install scikit-image``).
        :exc:`ValueError`: If an unvalid ``export_format`` is used.
    """

    try:
        import skimage  # noqa: F401
    except ImportError as e:
        logger.error("Please install `scikit-image` first: `pip install scikit-image`.")
        raise e

    print("Exporting dataset. This may take a while...")
    if export_format == ExportFormat.COCO_PANOPTIC:
        if dataset.task_type not in COMPATIBLE_TASK_TYPES[export_format]:
            raise ValueError(
                "Only datasets of type `segmentation-bitmap` and `segmentation-bitmap-highres` can be exported to this format."
            )
        from .export import export_coco_panoptic

        return export_coco_panoptic(dataset, export_folder)
    elif export_format == ExportFormat.COCO_INSTANCE:
        if dataset.task_type not in COMPATIBLE_TASK_TYPES[export_format]:
            raise ValueError(
                "Only datasets of type `segmentation-bitmap`, `segmentation-bitmap-highres`, `vector`, `bboxes` and `keypoints` can be exported to this format."
            )
        from .export import export_coco_instance

        return export_coco_instance(dataset, export_folder)
    elif export_format == ExportFormat.YOLO:
        if dataset.task_type not in COMPATIBLE_TASK_TYPES[export_format]:
            raise ValueError(
                "Only datasets of type `vector`, `bboxes` and `image-vector-sequence` can be exported to this format."
            )
        from .export import export_yolo

        return export_yolo(
            dataset,
            export_folder,
            image_width=kwargs.get("image_width", None),
            image_height=kwargs.get("image_height", None),
        )
    elif export_format in {
        ExportFormat.SEMANTIC_COLOR,
        ExportFormat.INSTANCE_COLOR,
        ExportFormat.SEMANTIC,
        ExportFormat.INSTANCE,
    }:
        if dataset.task_type not in COMPATIBLE_TASK_TYPES[export_format]:
            raise ValueError(
                "Only datasets of type `segmentation-bitmap` and `segmentation-bitmap-highres` can be exported to this format."
            )
        from .export import export_image

        return export_image(dataset, export_folder, export_format, id_increment)
    elif export_format == ExportFormat.POLYGON:
        if dataset.task_type not in COMPATIBLE_TASK_TYPES[export_format]:
            raise ValueError(
                "Only datasets of type `segmentation-bitmap` and `segmentation-bitmap-highres` can be exported to this format."
            )
        from .export import export_polygon

        return export_polygon(dataset, export_folder)
    else:
        raise ValueError("Please choose a valid `export_format`.")


def load_image_from_url(
    url: str,
    save_filename: Optional[str] = None,
    s3_client: Optional[Any] = None,
    gcp_client: Optional[Any] = None,
) -> Image.Image:
    """Load an image from url.

    Args:
        url: The image url.
        save_filename: The filename to save to.
        s3_client: A boto3 S3 client, e.g. ``s3_client = boto3.client("s3")``. Needs to be provided if your images are in a private S3 bucket. Defaults to :obj:`None`.
        gcp_client: A Google Gloud Storage client, e.g. ``gcp_client = storage.Client()``. Needs to be provided if your images are in a private GCP bucket. Defaults to :obj:`None`.

    Returns:
        A PIL image.
    """
    if s3_client is not None:
        url_parsed = urlparse(url)
        regex = re.search(r"(.+).(s3|s3-accelerate).(.+).amazonaws.com", url_parsed.netloc)
        if regex:
            bucket = regex.group(1)

            if bucket == "segmentsai-prod":
                image = Image.open(BytesIO(session.get(url).content))
            else:
                key = url_parsed.path.lstrip("/")

                file_byte_string = s3_client.get_object(Bucket=bucket, Key=key)["Body"].read()
                image = Image.open(BytesIO(file_byte_string))
    elif gcp_client is not None and url.startswith("gs://"):
        bucket_name, blob_name = url[5:].split("/", 1)
        bucket = gcp_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        file_byte_string = blob.download_as_bytes()
        image = Image.open(BytesIO(file_byte_string))
    else:
        image = Image.open(BytesIO(session.get(url).content))

    if save_filename is not None:
        if "exif" in image.info:
            image.save(save_filename, exif=image.info["exif"])
        else:
            image.save(save_filename)

    return image


def load_pointcloud_from_url(
    url: str,
    save_filename: Optional[str] = None,
    s3_client: Optional[Any] = None,
    gcp_client: Optional[Any] = None,
) -> o3d.geometry.PointCloud:
    """Load a point cloud from a url.

    Args:
        url: The point cloud url.
        save_filename: The filename to save to.
        s3_client: A boto3 S3 client, e.g. ``s3_client = boto3.client("s3")``. Needs to be provided if your point clouds are in a private S3 bucket. Defaults to :obj:`None`.
        gcp_client: A Google Gloud Storage client, e.g. ``gcp_client = storage.Client()``. Needs to be provided if your images are in a private GCP bucket. Defaults to :obj:`None`.

    Returns:
        A point cloud.

    Raises:
        :exc:`ImportError`: If open3d is not installed (to install run ``pip install open3d``).
    """

    try:
        import open3d as o3d
    except ImportError as e:
        logger.error("Please install open3d first: pip install open3d")
        raise e

    def load_pointcloud_from_parsed_url(url: str) -> o3d.geometry.PointCloud:
        with NamedTemporaryFile(suffix=".pcd") as f:
            f.write(session.get(url).content)
            pointcloud = o3d.io.read_point_cloud(f.name)

        return pointcloud

    if s3_client is not None:
        url_parsed = urlparse(url)
        regex = re.search(r"(.+).(s3|s3-accelerate).(.+).amazonaws.com", url_parsed.netloc)
        if regex:
            bucket = regex.group(1)

            if bucket == "segmentsai-prod":
                pointcloud = load_pointcloud_from_parsed_url(url)
            else:
                key = url_parsed.path.lstrip("/")
                file_byte_string = s3_client.get_object(Bucket=bucket, Key=key)["Body"].read()
                with NamedTemporaryFile(suffix=".pcd") as f:
                    f.write(file_byte_string)
                    pointcloud = o3d.io.read_point_cloud(f.name)
    elif gcp_client is not None and url.startswith("gs://"):
        bucket_name, blob_name = url[5:].split("/", 1)
        bucket = gcp_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        file_byte_string = blob.download_as_bytes()
        with NamedTemporaryFile(suffix=".pcd") as f:
            f.write(file_byte_string)
            pointcloud = o3d.io.read_point_cloud(f.name)
    else:
        pointcloud = load_pointcloud_from_parsed_url(url)
    if save_filename is not None:
        o3d.io.write_point_cloud(save_filename, pointcloud)

    return pointcloud


def load_label_bitmap_from_url(url: str, save_filename: Optional[str] = None) -> npt.NDArray[np.uint32]:
    """Load a label bitmap from url.

    Args:
        url: The label bitmap url.
        save_filename: The filename to save to.

    Returns:
        A :class:`numpy.ndarray` with :class:`numpy.uint32` ``dtype`` where each unique value represents an instance id.
    """

    def extract_bitmap(
        bitmap: Image.Image,
    ) -> npt.NDArray[np.uint32]:
        bitmap_array = np.array(bitmap)
        bitmap_array[:, :, 3] = 0
        bitmap_array = bitmap_array.view(np.uint32).squeeze(2)
        return bitmap_array

    bitmap = Image.open(BytesIO(session.get(url).content))
    bitmap_array = extract_bitmap(bitmap)

    if save_filename:
        Image.fromarray(bitmap_array).save(save_filename)

    return bitmap_array


def load_release(release: Release) -> Any:
    """Load JSON from Segments release.

    Args:
        release: A Segments release.

    Returns:
        A JSON with the release labels.
    """
    release_file = cast(str, release.attributes.url)  # TODO Fix in the backend.
    content = requests.get(release_file)
    return json.loads(content.content)


def handle_exif_rotation(image: Image.Image) -> Image.Image:
    """Handle the exif rotation of a PIL image.

    Args:
        image: A PIL image.

    Returns:
        A rotated PIL image.
    """

    def get_key_by_value(dictionary: Mapping[int, str], value: str) -> int:
        for k, v in dictionary.items():
            if v == value:
                return k
        raise ValueError(f"No such value {value}.")

    try:
        orientation = get_key_by_value(ExifTags.TAGS, "Orientation")
        exif = dict(image.getexif().items())
        if exif[orientation] == 3:
            image = image.transpose(Image.ROTATE_180)
        elif exif[orientation] == 6:
            image = image.transpose(Image.ROTATE_270)
        elif exif[orientation] == 8:
            image = image.transpose(Image.ROTATE_90)
        return image
    except (AttributeError, KeyError, IndexError, ValueError):
        return image


def show_polygons(
    image_directory_path: str,
    image_id: int,
    exported_polygons_path: str,
    seed: int = 0,
    output_path: Optional[str] = None,
) -> None:
    """Show the exported contours of a segmented image (i.e., resulting from :func:`.export_dataset` with polygon export format).

    Args:
        image_directory_path: The image directory path.
        image_id: The image id (this can be found in the exported polygons JSON file).
        exported_polygons_path: The exported polygons path.
        seed: The seed used to generate random colors. Defaults to ``0``.
        output_path: The directory to save the plot to. Defaults to :obj:`None`.

    Returns:
        None

    Raises:
        :exc:`ImportError`: If matplotlib is not installed.
    """

    try:
        from matplotlib import image
        from matplotlib import pyplot as plt
        from matplotlib.patches import Polygon
    except ImportError as e:
        logger.error("Please install `matplotlib` first: `pip install matplotlib`.")
        raise e

    def find_image_name(images: List[Dict[str, Any]], image_id: int) -> str:
        for image in images:
            if image["id"] == image_id:
                return cast(str, image["file_name"])
        raise KeyError("Cannot find the image id. Please provide a valid id.")

    def get_random_color() -> Tuple[float, float, float]:
        return (random.uniform(0, 1), random.uniform(0, 1), random.uniform(0, 1))

    def normalize(color: List[int]) -> Tuple[float, float, float]:
        """Transform a color from 0-255 range to 0-1 range and from a list to a tuple, e.g., `[255, 0, 123]` to `(1, 0, 0.5)`."""
        return (color[0] / 255, color[1] / 255, color[2] / 255)

    random.seed(seed)

    with open(exported_polygons_path, "r") as f:
        polygons = json.load(f)

    image_name = find_image_name(polygons["images"], image_id)
    image = image.imread(f"{image_directory_path}/{image_name}")

    # {category id: (category name, color)}
    categories = {
        category["id"]: (
            category["name"],
            normalize(category["color"]) if category["color"] else get_random_color(),
        )
        for category in polygons["categories"]
    }

    # {category id: polygons}
    annotations = defaultdict(list)
    filtered_annotations = filter(lambda dictionary: dictionary["image_id"] == image_id, polygons["annotations"])
    for annotation in filtered_annotations:
        annotations[annotation["category_id"]].extend(annotation["polygons"])

    # {category name: (polygons, color)}
    category_name_polygons_with_annotations = {
        category_name: (annotations[category_id], category_color)
        for category_id, (category_name, category_color) in categories.items()
        if annotations[category_id]
    }

    fig, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3, sharex=True, sharey=True, figsize=(25, 10))

    used_category_names = set()
    for category_name, (
        polygons,
        color,
    ) in category_name_polygons_with_annotations.items():
        for p in polygons:
            polygon = Polygon(
                xy=np.asarray(p).reshape(-1, 2),
                facecolor=color,
                edgecolor=color,
                label=category_name if category_name not in used_category_names else None,
                closed=True,
                alpha=0.5,
            )

            used_category_names.add(category_name)
            polygon_copy = copy.deepcopy(polygon)
            polygon_copy.set_label(None)

            ax1.add_patch(polygon)
            # An Artist, container or primitive, cannot be contained in multiple containers, which is consistent with the fact that each Artist holds the parent container as a bare object, not in a list.
            ax2.add_patch(polygon_copy)

    # Ax 2
    # ax2.axis("off")
    ax2.set_title("Both")
    ax2.imshow(image)
    ax2.set_xlabel("Width (pixels)")

    # Ax 1 (uses the aspect ratio of the image in axes 2)
    # ax1.axis("off")
    ax1.set_title("Label")
    # ax1.imshow(image)
    # https://stackoverflow.com/a/44655020
    aspect = np.diff(ax1.get_xlim())[0] / np.diff(ax1.get_ylim())[0]
    aspect /= np.diff(ax2.get_xlim())[0] / np.diff(ax2.get_ylim())[0]
    aspect = np.abs(aspect)
    ax1.set_aspect(aspect)
    ax1.set_xlabel("Width (pixels)")
    ax1.set_ylabel("Height (pixels)")

    # Ax 3
    # ax3.axis("off")
    ax3.set_title("Image")
    ax3.imshow(image)
    ax3.set_xlabel("Width (pixels)")

    fig.legend()

    if output_path:
        path = os.path.join(output_path, f"exported_polygons_from_image_id_{image_id:04d}")
        plt.savefig(path, bbox_inches="tight")

    plt.show()


def cuboid_to_segmentation(
    pointcloud: npt.NDArray[np.float32],
    label_attributes: PointcloudCuboidLabelAttributes,
    ego_pose: Optional[EgoPose] = None,
) -> npt.NDArray[np.uint32]:
    """Convert a cuboid label to an instance segmentation label.

    Args:
        pointcloud: A point cloud of size Nx3.
        label_attributes: A cuboid label from a single frame interface or one frame from a sequence interface.

    Returns:
        An instance segmentation label of size Nx1 mapping each point cloud point to a cuboid instance.

    Raises:
        :exc:`ImportError`: If pyquaternion is not installed (to install run ``pip install pyquaternion``).
        :exc:`ImportError`: If open3d is not installed (to install run ``pip install open3d``).
    """

    try:
        from pyquaternion import Quaternion
    except ImportError as e:
        logger.error("Please install pyquaternion first: pip install pyquaternion")
        raise e

    try:
        import open3d as o3d
    except ImportError as e:
        logger.error("Please install `open3d` first: `pip install open3d`")
        raise e

    # check dimensions of input
    assert pointcloud.shape[1] == 3, "Pointcloud must have shape (N, 3)"
    assert label_attributes.annotations, "Label must have annotations"

    # create cuboids
    cuboids = {}
    for annotation in label_attributes.annotations:
        center = np.array([annotation.position.x, annotation.position.y, annotation.position.z])
        extent = np.array([annotation.dimensions.x, annotation.dimensions.y, annotation.dimensions.z])
        if annotation.rotation:
            rotation = o3d.geometry.get_rotation_matrix_from_quaternion(
                np.array(
                    [
                        annotation.rotation.qx,
                        annotation.rotation.qy,
                        annotation.rotation.qz,
                        annotation.rotation.qw,
                    ]
                )
            )
        else:
            rotation = o3d.geometry.get_rotation_matrix_from_xyz((0, 0, annotation.yaw))

        # create cuboid
        cuboid = o3d.geometry.OrientedBoundingBox(center=center, extent=extent, R=rotation)

        cuboids[annotation.id] = cuboid

        # transform cuboids from world to lidar coordinate frame
        transformation = np.eye(4)
        if ego_pose and ego_pose.position:
            pos = ego_pose.position
            transformation[:3, 3] = np.array([pos.x, pos.y, pos.z])
        if ego_pose and ego_pose.heading:
            rot = ego_pose.heading
            rotq = Quaternion(x=rot.qx, y=rot.qy, z=rot.qz, w=rot.qw).inverse.transformation_matrix
            transformation[:3, :3] = rotq[:3, :3]
        # tranform = rotate + translate (bug: transform not defined for OrientedBoundingBox)
        cuboid.translate(-transformation[:3, 3])
        cuboid.rotate(transformation[:3, :3])

    # map each point to a cuboid instance
    result = np.zeros(pointcloud.shape[0], dtype=np.uint32)
    pointcloud = o3d.utility.Vector3dVector(pointcloud)
    for id, cuboid in cuboids.items():
        # get points inside cuboid
        inside = cuboid.get_point_indices_within_bounding_box(pointcloud)
        result[inside] = id

    return result


def array_to_pcd(
    positions: npt.NDArray[np.float32],
    output_path: str,
    intensity: Optional[npt.NDArray[np.float32]] = None,
    rgb: Optional[npt.NDArray[np.float32]] = None,
    compressed: bool = False,
    write_ascii: bool = True,
) -> None:
    """Convert a numpy array to a pcd file.

    Args:
        positions: Array of xyz points (Nx3 shape).
        output_path: Path to write the pcd.
        intensity: Optional array of intensity values (Nx1 shape).
        rgb: Optional array of rgb values (Nx3 shape) where red, green and blue are values between 0 and 255 or 0 and 1.
        compressed: If the pcd should be compressed. Defaults to :obj:`False`.
        write_ascii: If the pcd should be written in ascii format. Defaults to :obj:`True`.

    Returns:
        None

    Raises:
        :exc:`ImportError`: If open3d is not installed (to install run ``pip install open3d``).
        :exc:`AssertionError`: If the positions array does not have shape (N, 3).
        :exc:`AssertionError`: If the intensity array does not have shape (N, 1) or (N,).
        :exc:`AssertionError`: If the rgb array does not have shape (N, 3).
        :exc:`AssertionError`: If the intensity array does not have the same length as the positions array.
        :exc:`AssertionError`: If the rgb array does not have the same length as the positions array.
    """

    try:
        import open3d as o3d
    except ImportError as e:
        logger.error("Please install `open3d` first: `pip install open3d`")
        raise e

    assert positions.shape[1] == 3, f"Positions must have shape (N, 3) but has shape {positions.shape}"

    # cast to float32
    positions = positions.astype(np.float32)
    intensity = intensity.astype(np.float32) if intensity is not None else None
    rgb = rgb.astype(np.float32) if rgb is not None else None

    device = o3d.core.Device("CPU:0")
    dtype = o3d.core.float32
    pcd = o3d.t.geometry.PointCloud(device)
    pcd.point["positions"] = o3d.core.Tensor(positions, dtype, device)
    N = positions.shape[0]

    if intensity is not None:
        assert intensity.shape in (
            (N, 1),
            (N,),
        ), f"Intensity must have shape ({N}, 1) or ({N},) but has shape {intensity.shape}"
        if len(intensity.shape) == 1:
            intensity = intensity.reshape(-1, 1)

        pcd.point["intensity"] = o3d.core.Tensor(intensity, dtype, device)

    if rgb is not None:
        assert rgb.shape == (N, 3), f"RGB must have shape ({N}, 3) but has shape {rgb.shape}"

        # check rgb encoding (0-255 or 0-1)
        if np.max(rgb) > 1:
            rgb /= 255.0  # map 0-255 to 0-1 (open3d expects rgb values between 0 and 1)

        pcd.point["colors"] = o3d.core.Tensor(rgb, dtype, device)

    o3d.t.io.write_point_cloud(
        output_path,
        pcd,
        compressed=compressed,
        write_ascii=write_ascii,
        print_progress=True,
    )


def ply_to_pcd(ply_file: str, compressed: bool = False, write_ascii: bool = True) -> None:
    """Convert a .ply file to a .pcd file.

    Args:
        ply_file: The path to the .ply file.
        compressed: If the pcd should be compressed. Defaults to :obj:`False`.
        write_ascii: If the pcd should be written in ascii format. Defaults to :obj:`True`.

    Returns:
        None

    Raises:
        :exc:`ImportError`: If plyfile is not installed (to install run ``pip install plyfile``).
        :exc:`KeyError`: If the positions are not found in the ply file (expected colum names are ``x``, ``y`` and ``z``).
    """

    try:
        from plyfile import PlyData
    except ImportError as e:
        logger.error("Please install `plyfile` first: `pip install plyfile`")
        raise e

    with open(ply_file, "rb") as f:
        ply = PlyData.read(f)

    try:
        positions = np.stack((ply["vertex"]["x"], ply["vertex"]["y"], ply["vertex"]["z"]), axis=-1)
    except KeyError:
        raise KeyError("Could not find the positions in the ply file.")

    try:
        intensity = np.array(ply["vertex"]["intensity"]).reshape(-1, 1)
    except KeyError:
        try:
            intensity = np.array(ply["vertex"]["i"]).reshape(-1, 1)
        except KeyError:
            intensity = None

    try:
        rgb = np.stack((ply["vertex"]["r"], ply["vertex"]["g"], ply["vertex"]["b"]), axis=-1)
    except KeyError:
        try:
            rgb = np.stack(
                (ply["vertex"]["red"], ply["vertex"]["green"], ply["vertex"]["blue"]),
                axis=-1,
            )
        except KeyError:
            rgb = None

    pcd_extension = "_compressed.pcd" if compressed else ".pcd"
    pcd_path = ply_file.replace(".ply", pcd_extension)
    # prefer RGB over intensity (tiled point cloud does not support both)
    intensity = intensity if rgb is None else None
    array_to_pcd(
        positions,
        pcd_path,
        intensity=intensity,
        rgb=rgb,
        compressed=compressed,
        write_ascii=write_ascii,
    )


def las_to_pcd(las_file: str, compressed: bool = False, write_ascii: bool = True) -> None:
    """Convert a .las/.laz file to a .pcd file.

    Args:
        las_file: The path to the .las/.laz file.
        compressed: If the pcd should be compressed. Defaults to :obj:`False`.
        write_ascii: If the pcd should be written in ascii format. Defaults to :obj:`True`.

    Returns:
        None

    Raises:
        :exc:`ImportError`: If pdal is not installed (to install run ``pip install pdal``).
    """
    try:
        from pdal import Pipeline
    except ImportError:
        logger.error("Please install `pdal` first: `pip install pdal`")

    pipeline = json.dumps({"pipeline": [{"type": "readers.las", "filename": las_file}]})

    pipeline = Pipeline(pipeline)
    pipeline.execute()
    raw_array = pipeline.arrays[0]
    column_names = raw_array.dtype.names

    array = []
    for row in tqdm(raw_array, total=len(raw_array), color=SEGMENTS_ORANGE):
        array.append(np.array(list(row)))
    array = np.reshape(array, (-1, len(column_names)))

    def find_las_column_index(column_name: Union[str, List[str]]) -> Union[int, Tuple[int]]:
        def find_single_las_column_index(column_name: str) -> int:
            index = column_names.index(column_name)

            if index == -1:
                return ValueError(f"Can't find column index of column with name {column_name}")

            return index

        if isinstance(column_name, list):
            return tuple(find_single_las_column_index(column) for column in column_name)

        return find_single_las_column_index(column_name)

    def get_data(column_index: Union[int, List[int]]) -> np.ndarray:
        if isinstance(column_index, list):
            return np.stack((array[:, index] for index in column_index), axis=-1)

        return array[:, column_index]

    x, y, z = find_las_column_index(["X", "Y", "Z"])
    intensity = find_las_column_index("Intensity")
    r, g, b = find_las_column_index(["Red", "Green", "Blue"])

    positions = get_data([x, y, z])
    intensity = get_data(intensity)
    rgb = get_data([r, g, b])

    pcd_extension = "_compressed.pcd" if compressed else ".pcd"
    pcd_path = (
        las_file.replace(".las", pcd_extension)
        if las_file.endswith(".las")
        else las_file.replace(".laz", pcd_extension)
    )
    # prefer RGB over intensity (tiled point cloud does not support both)
    intensity = intensity if rgb is None else None
    array_to_pcd(
        positions,
        pcd_path,
        intensity=intensity,
        rgb=rgb,
        compressed=compressed,
        write_ascii=write_ascii,
    )


def sample_pcd(pcd_path: str, points: int = 500_000, output_path: Optional[str] = None) -> None:
    """Sample a point cloud to a given number of points.

    Args:
        pcd_path: The path to the point cloud.
        points: The number of points to sample. Defaults to ``500_000``.
        output_path: The path to save the sampled point cloud to. Defaults to :obj:`None`.

    Returns:
        None

    Raises:
        :exc:`ImportError`: If open3d is not installed (to install run ``pip install open3d``).
    """

    try:
        import open3d as o3d
    except ImportError as e:
        logger.error("Please install `open3d` first: `pip install open3d`")
        raise e

    if output_path is None:
        output_path = output_path.replace(".pcd", "_sampled.pcd")

    pcd = o3d.io.read_point_cloud(pcd_path)
    # open3d expects a step size (not a number of points)
    points_step_size = len(pcd.points) // points
    pcd = pcd.uniform_down_sample(points_step_size)
    o3d.io.write_point_cloud(output_path, pcd, write_ascii=False, compressed=True, print_progress=True)


def find_camera_rotation(client: SegmentsClient, dataset_identifier: str) -> Quaternion:
    """Find the correct camera rotation by trying all possibilities.

    Args:
        client: A Segments client.
        dataset_identifier: The dataset identifier.

    Returns:
        A :class:`pyquaternion.Quaternion` representing the correct rotation.

    Raises:
        :exc:`ImportError`: If pyquaternion is not installed (to install run ``pip install pyquaternion``).
        :exc:`ValueError`: If the dataset is not a point cloud sequence dataset.
        :exc:`ValueError`: If the user answers neither `y(es)` nor `n(o)` (case insensitive).
        :exc:`ValueError`: If the correct rotation is not found.
        :exc:`AlreadyExistsError`: If the cloned dataset already exists.
    """

    try:
        from pyquaternion import Quaternion
    except ImportError as e:
        logger.error("Please install `pyquaternion` first: `pip install pyquaternion`")
        raise e

    # 6 options for x axis, 4 options for y axis and 1 option for z axis
    X_AXIS_ROTATIONS = [
        Quaternion(axis=[0, 0, 1], angle=0),
        Quaternion(axis=[0, 0, 1], angle=np.pi / 2),
        Quaternion(axis=[0, 0, 1], angle=np.pi),
        Quaternion(axis=[0, 0, 1], angle=3 * np.pi / 2),
        Quaternion(axis=[0, 1, 0], angle=np.pi / 2),
        Quaternion(axis=[0, 1, 0], angle=3 * np.pi / 2),
    ]
    Y_AXIS_ROTATIONS = [
        Quaternion(axis=[1, 0, 0], angle=0),
        Quaternion(axis=[1, 0, 0], angle=np.pi / 2),
        Quaternion(axis=[1, 0, 0], angle=np.pi),
        Quaternion(axis=[1, 0, 0], angle=3 * np.pi / 2),
    ]
    # inverse rotation is possible
    total_rotation_options = len(X_AXIS_ROTATIONS) * len(Y_AXIS_ROTATIONS) * 2

    # clone dataset
    cloned_dataset_owner, cloned_dataset_name = (
        dataset_identifier.split("/")[0],
        f"{dataset_identifier.split('/')[-1]}-find-camera-rotation",
    )
    try:
        cloned_dataset = client.clone_dataset(dataset_identifier, cloned_dataset_name)
    except AlreadyExistsError:
        client.delete_dataset(f"{cloned_dataset_owner}/{cloned_dataset_name}")
        cloned_dataset = client.clone_dataset(dataset_identifier, cloned_dataset_name)

    # get (cloned) samples
    samples = client.get_samples(dataset_identifier)
    cloned_samples = client.get_samples(cloned_dataset.full_name)
    if not isinstance(samples[0].attributes, PointcloudSequenceSampleAttributes):
        raise ValueError(
            "Brute force camera rotations only works for point cloud sequence datasets. Reach out to `support@segments.ai` and `arnaud@segments.ai` if you are interested in this functionality for other dataset types."
        )

    # rotate images
    for xi, x_rot in enumerate(X_AXIS_ROTATIONS):
        for yi, y_rot in enumerate(Y_AXIS_ROTATIONS):
            for invert in [False, True]:
                for sample, cloned_sample in zip(samples, cloned_samples):
                    for frame, cloned_frame in zip(sample.attributes.frames, cloned_sample.attributes.frames):
                        for image, cloned_image in zip(frame.images, cloned_frame.images):
                            if not image.extrinsics or not image.extrinsics.rotation:
                                continue

                            rot = image.extrinsics.rotation
                            rot_q = Quaternion(x=rot.qx, y=rot.qy, z=rot.qz, w=rot.qw)

                            if invert:
                                rot_q = rot_q.inverse

                            new_rot_q = x_rot * y_rot * rot_q
                            new_rot = XYZW(
                                qx=new_rot_q.x,
                                qy=new_rot_q.y,
                                qz=new_rot_q.z,
                                qw=new_rot_q.w,
                            )
                            cloned_image.extrinsics.rotation = new_rot
                    client.update_sample(cloned_sample.uuid, attributes=cloned_sample.attributes)

                offset, y_offset, x_offset = 1, 2, 2 * len(Y_AXIS_ROTATIONS)
                rotation_option = xi * x_offset + yi * y_offset + invert + offset

                rotation_OK = input(
                    f"Correct image rotation in {cloned_dataset.full_name} (rotation {rotation_option}/{total_rotation_options})? [y/n]"
                )
                if rotation_OK.lower() in ["y", "yes"]:
                    client.delete_dataset(cloned_dataset.full_name)
                    return (x_rot * y_rot).inverse if invert else x_rot * y_rot
                elif rotation_OK.lower() not in ["n", "no"]:
                    raise ValueError(f"Please enter `y(es)` or `n(o)` (case insensitive) not `{rotation_OK}`.")

    raise ValueError("Correct rotation not found.")


def add_kinematics_to_label(
    sample: Sample,
    label: Label,
    timestamp_unit: Optional[int] = 1_000_000_000.0,
    fps: Optional[float] = None,
) -> Label:
    """Add velocity and acceleration calculations to point cloud sequence cuboid labels.

    This function calculates velocity and acceleration (relative and absolute) for annotations with position, ego_position and timestamp data across frames in point cloud sequence datasets. If timestamp data is missing, it estimates a frame per 0.5 seconds. If ego-position data is missing, only absolute kinematics are calculated.

    Args:
        sample: A Sample object containing the point cloud sequence sample data.
        label: A Label object containing point cloud sequence data.
        timestamp_unit: Conversion factor for timestamps to seconds. Defaults to ``1_000_000_000.0`` (i.e., timestamps in nanoseconds).
        fps: Frames per second to use when timestamps are missing.

    Returns:
        A copy of the label with velocity and acceleration attributes added to annotations.

    Raises:
        ValueError: If the sample is not a multisensor or point cloud sequence cuboid label.
        ValueError: If the label sample UUID does not match the sample UUID.
        ValueError: If fps is not provided when timestamps are missing.

    Example:

      .. code-block:: python

        # pip install segments-ai
        from segments import SegmentsClient
        from segments.utils import add_kinematics_to_label

        client = SegmentsClient('YOUR_API_KEY')
        sample = client.get_sample(sample_uuid)
        label = client.get_label(sample_uuid)
        updated_label = add_kinematics_to_label(sample, label)
    """

    def _calculate_velocity_and_acceleration(
        ego_positions: List[Optional[npt.NDArray[np.float64]]],
        positions: List[Optional[npt.NDArray[np.float64]]],
        timestamps: List[Optional[float]],
    ) -> Tuple[List[Optional[float]], List[Optional[float]]]:
        velocities: List[Optional[Velocity]] = []
        accelerations: List[Optional[Acceleration]] = []

        previous_ego_position = None
        previous_position = None
        previous_timestamp = None
        previous_index = None
        previous_velocity_vector = None
        previous_ego_velocity_vector = None

        for i, (ego_position, position, timestamp) in enumerate(zip(ego_positions, positions, timestamps)):
            if previous_position is None or position is None:
                velocities.append(None)
                accelerations.append(None)

                if position is not None:
                    previous_ego_position = ego_position
                    previous_position = position
                    previous_timestamp = timestamp
                    previous_index = i

                continue

            velocity = None
            acceleration = None

            time_delta = None
            if timestamp is None or previous_timestamp is None:
                if fps is None:
                    raise ValueError("FPS must be provided if timestamps are missing.")
                frame_index_delta = i - previous_index
                time_delta = frame_index_delta / fps
            else:
                time_delta = (timestamp - previous_timestamp) / timestamp_unit
            if time_delta <= 0:
                raise ValueError("Timestamps must be increasing.")

            velocity_vector = (position - previous_position) / time_delta

            ego_velocity_vector = None
            relative_velocity = None
            if ego_position is not None and previous_ego_position is not None:
                ego_velocity_vector = (ego_position - previous_ego_position) / time_delta
                relative_velocity = float(np.linalg.norm(velocity_vector - ego_velocity_vector))

            velocity = Velocity(
                absolute=float(np.linalg.norm(velocity_vector)),
                relative=relative_velocity,
            )

            if previous_velocity_vector is not None:
                acceleration_vector = (velocity_vector - previous_velocity_vector) / time_delta

                relative_acceleration = None
                if ego_velocity_vector is not None and previous_ego_velocity_vector is not None:
                    ego_acceleration_vector = (ego_velocity_vector - previous_ego_velocity_vector) / time_delta
                    relative_acceleration = float(np.linalg.norm(acceleration_vector - ego_acceleration_vector))

                acceleration = Acceleration(
                    absolute=float(np.linalg.norm(acceleration_vector)),
                    relative=relative_acceleration,
                )

            previous_velocity_vector = velocity_vector
            previous_ego_velocity_vector = ego_velocity_vector

            previous_ego_position = ego_position
            previous_position = position
            previous_timestamp = timestamp
            previous_index = i

            velocities.append(velocity)
            accelerations.append(acceleration)

        return velocities, accelerations

    if label.sample_uuid != sample.uuid:
        raise ValueError("Label sample UUID does not match sample UUID.")

    if isinstance(sample.attributes, PointcloudSequenceSampleAttributes):
        sample_pointcloud_frames = sample.attributes.frames
    elif isinstance(sample.attributes, MultiSensorSampleAttributes):
        pointcloud_sensor = None
        for sensor in sample.attributes.sensors:
            if hasattr(sensor, "task_type") and sensor.task_type == TaskType.POINTCLOUD_CUBOID_SEQUENCE:
                pointcloud_sensor = sensor
                break

        if pointcloud_sensor is None:
            raise ValueError(
                "Sample must contain a pointcloud cuboid sensor to calculate kinematics. "
                "No sensor with task_type 'pointcloud-cuboid-sequence' found in sample."
            )
        sample_pointcloud_frames = pointcloud_sensor.attributes.frames
    else:
        raise ValueError(
            "Sample must be a pointcloud cuboid sequence or multisensor sample to calculate kinematics. "
            "Supported types: PointcloudSequenceSample, MultiSensorSample."
        )

    label_copy = copy.deepcopy(label)
    sequence_attributes = []

    if isinstance(label_copy.attributes, PointcloudSequenceCuboidLabelAttributes):
        if hasattr(label_copy.attributes, "frames") and label_copy.attributes.frames:
            sequence_attributes.append(label_copy.attributes)
    elif isinstance(label_copy.attributes, MultiSensorLabelAttributes):
        for sensor in label_copy.attributes.sensors:
            if isinstance(
                sensor,
                MultiSensorPointcloudSequenceCuboidLabelAttributes,
            ):
                if hasattr(sensor.attributes, "frames") and sensor.attributes.frames:
                    sequence_attributes.append(sensor.attributes)

    if not sequence_attributes:
        raise ValueError(
            "Label must be a point cloud sequence type with frames to calculate kinematics. "
            "Supported label types: PointcloudSequenceCuboidLabel, "
            "or MultiSensorLabel containing point cloud sequence sensors."
        )

    ego_positions = []
    for frames in sample_pointcloud_frames:
        if not hasattr(frames, "ego_pose") or frames.ego_pose is None or not hasattr(frames.ego_pose, "position"):
            ego_positions.append(None)
        else:
            pos = frames.ego_pose.position
            ego_positions.append(np.array([pos.x, pos.y, pos.z]))

    for seq_attr in sequence_attributes:
        tracks: Dict[int, List[Tuple[int, Any]]] = defaultdict(list)

        for frame_idx, frame in enumerate(seq_attr.frames):
            if not hasattr(frame, "annotations"):
                continue

            for annotation in frame.annotations:
                if not hasattr(annotation, "track_id") or not hasattr(annotation, "position"):
                    continue
                tracks[annotation.track_id].append((frame_idx, annotation))

        for track_annotations in tracks.values():
            track_annotations.sort(key=lambda x: x[0])

            positions = []
            timestamps = []
            annotation_refs = []

            for frame_idx, annotation in track_annotations:
                frame = seq_attr.frames[frame_idx]

                position = None
                if annotation.position is not None:
                    position = np.array([annotation.position.x, annotation.position.y, annotation.position.z])

                timestamp = frame.timestamp
                if isinstance(timestamp, str):
                    try:
                        timestamp = float(timestamp)
                    except ValueError:
                        timestamp = None
                elif isinstance(timestamp, (int, float)):
                    timestamp = float(timestamp)

                timestamps.append(timestamp)
                positions.append(position)
                annotation_refs.append(annotation)

            velocities, accelerations = _calculate_velocity_and_acceleration(ego_positions, positions, timestamps)

            for i, (velocity, acceleration) in enumerate(zip(velocities, accelerations)):
                annotation = annotation_refs[i]
                if velocity is not None:
                    annotation.velocity = velocity

                if acceleration is not None:
                    annotation.acceleration = acceleration

    return label_copy
